"""Escher JSON-schema objects

This modules has all the JSON objects that can be seen in the original JSON
schema for Escher.

The classes retain their original JSON names and attributes.
- Node: Represents the nodes. They can be a metabolite or markers. They give
the position of all dots in the canvas.
- Segment: Represents the connections between nodes.
- Reaction: Represents a reaction. They contain the needed segments for the
visualization of the pathway.
"""
from contextlib import suppress
from collections import UserDict
from typing import Any

from cobramod.error import NodeAttributeError
from cobramod.visualization.pair import PairDictionary
from cobramod.visualization.debug import debug_log


class Node(UserDict):
    """
    Simple class that represent a node for the JSON schema of Escher. The
    node type can be either a 'metabolite', 'midmarker' or a 'multimarker'.
    Markers only need the arguments x and y. Check method
    :func:`cobramod.visualization.items.Node.__init__` to see all the
    corresponding arguments.
    """

    def __init__(
        self: Any,
        node_type: str,
        x: float,
        y: float,
        label_x: float = None,
        label_y: float = None,
        bigg_id: str = "",
        name: str = "",
        node_is_primary: bool = False,
    ):
        """
        Build a node for JsonCobramod, which can be parsed later into a proper
        JSON for Escher. A node can be a metabolite, midmarker or multimarker.
        Markers are the visible dots that are located in the middle of the
        reaction that defines the orientation (vertical, horizonzal).
        Metabolites represent the orange dots.

        Args:
            node_type: Can either be a 'metabolite', 'midmarker' or
                'multimarker'. Metabolites must have all arguments. Markers
                only need y and x.
            x (float): Location in x-axis for the node.
            y (float): Location in y-axis for the node
            label_x (float): Location in x-axis for the label of the node.
            label_y (float): Location in y-axis for the label of the node.
            bigg_id (str, optional): Identifier for the metabolite. It does not
                have to be a bigg identifier.
            name (str, optional): The name for the metabolite.
            node_is_primary (bool, optional): If the metabolite is a primary
                compound.
        """
        # Using a copy of locals, without self and the kwargs
        kwargs = {
            key: value
            for key, value in locals().copy().items()
            if key not in ("kwargs", "__class__", "self")
        }
        for method in (self.metabolite, self.marker):
            with suppress(NodeAttributeError):
                method(**kwargs)
        if "node_type" not in self.data.keys():
            raise NodeAttributeError(
                "Wrong attributes. Check 'node_types' and fix the arguments."
            )

    def marker(self, node_type: str, x: float, y: float, **kwargs):
        """
        Sets the attribute of the dictionary if the dictionary belongs to a
        midmarker or multimarker. Arguments specified in __init__.
        """
        # The kwargs is just to get rid of arguments that do not belong here
        # Must be yield, otherwise, the for loop wont work
        if node_type in ("midmarker", "multimarker"):
            super().__init__(self, node_type=node_type, x=x, y=y)
        else:
            raise NodeAttributeError(
                "Given node type does not represent a marker"
            )

    def metabolite(
        self,
        node_type: str,
        x: float,
        y: float,
        label_x: float,
        label_y: float,
        bigg_id: str = "",
        name: str = "",
        node_is_primary: bool = False,
    ):
        """
        Returns a dictionary with the arguments if node_type correspond to
        metabolite. Arguments specified in __init__,
        """
        # Must be yield, otherwise, the for loop wont work
        if node_type == "metabolite":
            kwargs = {
                key: value
                for key, value in locals().copy().items()
                if key not in ("kwargs", "__class__", "self")
            }
            super().__init__(self, **kwargs)
        else:
            raise NodeAttributeError(
                "Given node type does not represent a metabolite"
            )


class Segment(UserDict):
    """
    Build a segment for th reaction object for JsonCobramod, which can be
    later parsed into a proper JSON for Escher. A segment needs a starting
    position and and end-position. Additionally, the dots for the
    modification of the arrows (b1, b2) can be included.

    Keyword Arguments:
        from_node_id (str): Identifier for the starting node.
        to_node_id (str): identifier for the end node.
        b1 (dict, optional): dictionary with x and y for the blue dot that
            shapes the arrow.
        b2 (dict, optional): dictionary with x and y for the blue dot that
        shapes the arrow.
    """

    def __init__(self, *args, **kwargs):
        """
        Build a segment for the reactionn object that will be included in
        JsonCobramod. Check, documentation of the class for arguments of this
        method.
        """
        super().__init__(self, *args, **kwargs)
        for key in ("from_node_id", "to_node_id", "b1", "b2"):
            try:
                self.data[key]
            except KeyError:
                if key in ("from_node_id", "to_node_id"):
                    raise ValueError(f"Argument '{key}' missing")
                elif key in ("b1", "b2"):
                    self.data[key] = None


class Reaction(UserDict):
    """
    Simple class that represent a reaction for the JSON schema for Escher. A
    Reaction have the Segment that give the position of the reaction in the
    canvas.

    Keyword Arguments:
        name (str): Name of the reaction.
        bigg_id (str): Identifier of the reactions. It does not have to be
            from the database BIGG.
        reversibility (bool): If reaction is reversible
        label_y (float): Position in y-axis for the label.
        label_x (float): Position in x-axis for the label.
        gene_reaction_rule (str, optional): Gene rules that specify
            involved genes.
        genes (list, optional): A list with the genes involved for that
            identifier. Each item has to be a dictionary with the keys
            'bigg_id' and 'name'
        metabolites (list, optional): A list with the metabolites involved.
            Each iteam has to be an dictionary with the keys 'bigg_id' and
            'coefficient'.
        segments (PairDictionary, optional): dictionary with
            :func:`cobramod.visualization.items.Segment`. They represent
            the connections between nodes.
    """

    def __init__(self, *args, **kwargs):
        """
        Creates an object with the information for the representation of a
        reaction in Escher. All keyword arguments are located in the docstring
        of the class
        """
        super().__init__(self, *args, **kwargs)
        for key in (
            "name",
            "bigg_id",
            "reversibility",
            "label_x",
            "label_y",
            "gene_reaction_rule",
            "genes",
            "metabolites",
            "segments",
        ):
            try:
                self.data[key]
            except KeyError:
                # Arguments must be passed this way, otherway UserDict loses
                # its functionalities.
                if key in (
                    "name",
                    "bigg_id",
                    "reversibility",
                    "label_x",
                    "label_y",
                ):
                    # Obligatory arguments
                    raise ValueError(f"Argument '{key}' missing")
                elif key == "gene_reaction_rule":
                    self.data[key] = ""
                elif key in ("genes", "metabolites"):
                    self.data[key] = list()
                elif key == "segments":
                    self.data[key] = PairDictionary()
        debug_log.debug(f'New Reaction "{self.data["bigg_id"]}" created.')

    def add_metabolite(self, bigg_id: str, coefficient: float):
        """
        Add the identifier for a metabolite and its corresponding coefficient
        into the 'metabolite' key.

        Args:
            bigg_id (str): identifier of the metabolite. It does not have to be
                from BIGG.
            coefficient (float): Coefficient of the metabolite for the
                reaction.
        """
        self.data["metabolites"].append(
            dict(bigg_id=bigg_id, coefficient=coefficient)
        )
        debug_log.debug(
            f"Metabolite information added to Reaction "
            f'"{self.data["bigg_id"]}"'
            f' with id "{bigg_id}". Coefficient: {coefficient}.'
        )

    def add_segment(self, identifier: str, from_node_id: str, to_node_id: str):
        """
        Adds to the Reaction for given identifier a Segment. It will be
        visualized in Escher.

        Args:
            identifier (str): Number-identifier for the Segment,
            from_node_id (str): json data identifier, that represents the
                initial node for the segment.
            to_node_id (str): json data identifier, that represents the last
                node for the segment.
        """
        self.data["segments"].update(
            {
                identifier: Segment(
                    from_node_id=from_node_id, to_node_id=to_node_id
                )
            }
        )
        debug_log.debug(
            f'New Segment "{self.data["bigg_id"]}" in Reaction '
            f'"{identifier}" From: {from_node_id}, to {to_node_id}'
        )



"""COBRApy Group-child class extension

The new class :class:`cobramod.pathway.Pathway" is child derived from
:class:`cobra.core.group.Group`. It extends some functionalities such as:

- solution: Obtain the solution for the specific members.
- visualize: get a :class:`escher.Builder` for that specific Pathway.
"""
from __future__ import annotations

import warnings
from pathlib import Path
from typing import Any, Optional, Union

import cobra.core as cobra_core

try:
    import escher

    _has_escher = True
except ImportError:
    _has_escher = False

import pandas as pd

from cobramod.debug import debug_log
from cobramod.error import GraphKeyError
from cobramod.visualization.converter import JsonDictionary


class Pathway(cobra_core.Group):
    """
    A Sub-class from the original COBRApy :class:`cobra.Group`, which inherits
    all attributes and adds the method solution, to get a Solution for the
    members of this Class.

    Attributes
        vertical (bool, optional):
            Variable that determines whether the display should be vertical or
            horizontal using Escher. Defaults to False.
        color_negative (str or list of int, optional) :
            The color to use as the endpoint for the negative fluxes during the
            creation of the color gradient. All colors of the CSS standard can
            be used here or their RGB representation.
        color_positive (str or list of int, optional) :
            The color to use as the endpoint for the positive fluxes during the
            creation of the color gradient. All colors of the CSS standard can
            be used here or their RGB representation.
        color_min_max (list of float, optional) :
            The maximum and minimum to be taken into account when creating the
            color gradient. This creates these two values artificially to allow
            the creation of a data-independent color gradient. Fluxes larger or
            smaller are ignored accordingly.
        color_quantile (bool, optional) :
            Attribute that defines whether the color gradient should be
            determined through quantiles or equally distributed between the
            maximum and the minimum. Defaults to False which means that the
            gradations are evenly distributed.
        color_n_steps (int, optional) :
            The number of steps used when creating the color gradient. Uses the
            number of fluxes by default. The default value is None.
        color_max_steps (int, optional) :
            The maximum number of steps to use when creating the color
            gradient. Default value is 100.
    See Also:
            Color names according to the css standard:
            https://www.w3schools.com/cssref/css_colors.asp
    """

    def __init__(
        self,
        id: str,
        name: str = "",
        members: Optional[set[cobra_core.Object]] = None,
        kind: Optional[str] = None,
    ):
        """
        Regular __init__ of the class. It maintains the same attributes as its
        parent :class:`cobra.core.group.Group`. Included is the order of the
        reactions.

        Args:
            id (str): Identifier of the pathway.
            name (str): Name for the pathway.
            members (set): Includes the members for the pathway. It should
                include only :class:`cobra.core.reaction.Reaction`. Defaults to
                None.
            kind (str): The kind of group, as specified for the Groups feature
                in the SBML level 3 package specification. Can be any of
                "classification", "partonomy", or "collection". The default is
                "collection". Please consult the SBML level 3 package
                specification to ensure you are using the proper value for
                kind. (Text extracted from parent.)
        """
        super().__init__(
            id=id,
            name=name,
            kind=kind,
        )

        # Loop has to be after __init__, otherwise, behavior of class changes.
        self.graph = dict()
        self.notes: dict[str, Any] = {"ORDER": dict()}

        if members:
            for member in members:
                # Remove no Reactions
                if not isinstance(member, cobra_core.Reaction):
                    debug_log.debug(
                        f'Member "{member.id}" is not a Reaction. This object will'
                        f'not be included in the pathway "{self.id}".'
                    )
                    continue
                super().add_members(new_members=[member])

        # Attributes that can be used for the customization of the
        # visualization
        self.color_negative: Optional[Union[str, list[int]]] = None
        self.color_positive: Optional[Union[str, list[int]]] = None
        self.color_min_max: Optional[list[float]] = None
        self.color_max_steps: int = 100
        self.color_n_steps: Optional[int] = None
        self.color_quantile: bool = False
        self.vertical: bool = False

    def _filter(
        self, solution: cobra_core.Solution, attribute: str
    ) -> pd.Series:
        """
        Filters given attribute of Solution. Only members of the Pathway will
        appear.
        """
        return getattr(solution, attribute).filter(
            items=[member.id for member in self.members], axis="index"
        )

    def add_members(self, new_members: list[cobra_core.Reaction]):
        """
        Add given list of :class:`cobra.core.reaction.Reaction` into the
        Pathway.

        Args:
            new_members (list): List of Reactions to add to the class.

        Raises:
            TypeError: If not all members are proper Reaction objects.
        """
        if not all(
            [isinstance(member, cobra_core.Reaction) for member in new_members]
        ):
            raise TypeError("Not all given members are Reactions.")

        super().add_members(new_members=new_members)
        for member in new_members:
            try:
                # There is no need to extend the tuple or strings because
                # the father-reaction has at least 1 reaction
                self.notes["ORDER"][member.id]
            except KeyError:
                self.notes["ORDER"][member.id] = None

    def solution(self, solution: cobra_core.Solution) -> cobra_core.Solution:
        """
        Returns a :class:`cobra.Solution` with only the members of the pathway.

        Args:
            solution (Solution): Original COBRApy :class:`cobra.Solution` to
                filter.
        Returns:
            Solution: Filtered solution containing only members of the Pathway
                class.
        """
        return cobra_core.Solution(
            objective_value=solution.objective_value,
            status=solution.status,
            fluxes=self._filter(solution=solution, attribute="fluxes"),
            reduced_costs=self._filter(
                solution=solution, attribute="reduced_costs"
            ),
            shadow_prices=self._filter(
                solution=solution, attribute="shadow_prices"
            ),
        )

    @classmethod
    def _transform(cls, obj: cobra_core.Group) -> "Pathway":
        """
        Transform given :class:`cobra.Group` into a proper cobramod Pathway
        object.
        """
        pathway = cls(
            id=obj.id, name=obj.name, members=obj.members, kind=obj.kind
        )
        pathway._model = obj._model
        # Add directly from the notes or create a quick order
        try:
            # When loading model, COBRApy will modify the string
            if isinstance(obj.notes["ORDER"], str):
                obj.notes["ORDER"] = eval(
                    obj.notes["ORDER"].replace("&apos;", '"')
                )
            pathway.graph = pathway.notes["ORDER"] = obj.notes["ORDER"]

        except KeyError:
            # Check only Reactions
            reactions = [
                rxn
                for rxn in obj.members
                if isinstance(rxn, cobra_core.Reaction)
            ]
            order = {}
            for value, reaction in enumerate(reactions):
                try:
                    order[reaction.id] = reactions[value + 1].id
                except IndexError:
                    # It should be the last
                    order[reaction.id] = None
            pathway.graph = pathway.notes["ORDER"] = order

        debug_log.info(
            f'Group-object "{pathway.id}" was transformed to a Pathway-object.'
        )
        return pathway

    def modify_graph(self, reaction: str, next_reaction: Union[str, None]):
        """
        Modifies the order of the graph. This is useful when merging multiple
        pathways or joining reactions. In the graph, the selected reaction
        will be forced to show "next_reaction" as its successor.

        Args:
            reaction (str): Identifier of the reaction to modify in the graph.
            next_reaction (str, None): Identifier of the next reaction. This
                reaction will take place after "reaction". If None is passed,
                then "reaction" will not have successors.

        Raises:
            GraphKeyError: If the reaction or the next_reaction does not appear
            in the graph of the pathway.
        """
        # FIXME: add behavior for changing not NoneTypes
        # Pathway.graph is responsable for the mapping
        if next_reaction not in self.graph.keys() and next_reaction:
            raise GraphKeyError(
                f'Pathway "{self.id}" does not have a reaction '
                + f'{next_reaction}". Check that the reaction exist.'
            )
        try:
            self.graph[reaction]
            self.graph[reaction] = next_reaction
            # We need to modify the notes as well because it will change the
            # notes
            msg = f'The reaction order of pathway "{self.id}" was modified. '
            if not next_reaction:
                msg += f'There is no next reaction after "{reaction}".'
            else:
                msg += (
                    f'Reaction "{next_reaction}" takes place after '
                    + f'"{reaction}".'
                )
            self.notes["ORDER"][reaction] = next_reaction
            debug_log.info(msg=msg)
        except KeyError:
            raise GraphKeyError(
                f'Pathway "{self.id}" does not have a reaction '
                + f'{reaction}". Check that the reaction exist.'
            )

    def visualize(
        self,
        solution_fluxes: Optional[
            Union[cobra_core.Solution, dict[str, float]]
        ] = None,
        filename: Optional[Union[str, Path]] = None,
    ) -> Optional[escher.Builder]:
        """
        Returns a :class:`escher.Builder`, which can be used to create visual
        representations of the pathway.

        Args:
            solution_fluxes (Solution, dict): Series or Dictionary with fluxes.
                The values will be then showed in the Builder.
                Defaults to None.
            filename (str, Path): Path for the HTML. Defaults to
                "pathway.html" in the current working directory.
        """
        json_dict = JsonDictionary()
        if filename is None:
            filename = "pathway.html"

        # Define solution. If None, nothing will be added. Either dict or
        # regular solution
        if solution_fluxes is not None:
            if isinstance(solution_fluxes, cobra_core.Solution):
                solution_fluxes = solution_fluxes.fluxes.to_dict()
            json_dict.flux_solution = solution_fluxes

        # Get graph and add to json_dict
        json_dict.graph = self.graph.copy()
        reactions: dict[str, str] = {m.id: m.reaction for m in self.members}
        json_dict.reaction_strings = reactions

        if _has_escher:
            builder = json_dict.visualize(
                filepath=filename,
                vertical=self.vertical,
                color=[self.color_positive, self.color_negative],
                min_max=self.color_min_max,
                quantile=self.color_quantile,
                max_steps=self.color_max_steps,
                n_steps=self.color_n_steps,
            )
            debug_log.info(f'Visualization saved in "{filename}"')
            return builder
        else:
            warnings.warn(
                "Package Escher was not found. No visualization in available. "
                "You can install this extra-dependency running "
                "'pip install cobramod[escher]'"
            )
            return

    def _repr_html_(self):
        """
        Returns a HTML string with the attributes of the Pathway
        """
        return f"""
<table> <tbody> <tr> <td><strong>Pathway identifier</strong></td>
<td>{self.id}</td> </tr> <tr> <td><strong>Name</strong></td>
<td>{self.name}</td> </tr> <tr> <td><strong>Memory address</strong></td>
<td>0x0{id(self)}</td> </tr> <tr> <td><strong>Reactions involved</strong></td>
<td> <p>{", ".join([rxn.id for rxn in self.members])}</p> </td> </tr> <tr>
<td><strong>Genes involved<br /></strong></td> <td> <p>{", ".join([gene.id for
rxn in self.members for gene in rxn.genes])}</p> </td> </tr> <tr>
<td><strong>Visualization attributes</strong></td> <td> <ul> <li>vertical =
{self.vertical}</li> <li>color_negative = {self.color_negative}</li>
<li>color_positive = {self.color_positive}</li> <li>color_quantile =
{self.color_quantile}</li> </ul> </td> </tr> </tbody> </table> <p>&nbsp;</p>"""


def model_convert(model: cobra_core.Model):
    """
    Converts all Group objects in the given model to proper cobramod
    :class:`cobramod.pathway.Pathway`
    """
    for index, group in enumerate(iterable=model.groups):
        model.groups[index] = Pathway._transform(obj=group)



"""Core package

This sub-package contains the principal modules, which are responsible for the
creation of COBRApy objects and the addition of Pathway objects into the
metabolic model. The modules include:

- creation: Creation of COBRApy objects
- crossreferences: Storing and retriving crossreferences information to COBRApy
  objects
- extension: Addition of reactions and metabolites into metabolic models
- genes: Creation of COBRApy genes
- graph: Algorithm to build the order of reactions
- pathway: Creation of :func:`cobra.core.group.Group` child.
:func:`cobramod.core.pathway.Pathway`.
- summary: Short and long information about changes
"""



#!/usr/bin/env python3
import tempfile
import unittest
from logging import DEBUG
from pathlib import Path
from unittest import TestCase
from unittest.mock import patch

import pandas as pd
from cobra import Metabolite, Reaction
from cobra import __version__ as cobra_version
from cobramod import __version__ as cmod_version
from cobramod.core.crossreferences import (
    add2dict_unique,
    add_crossreferences,
    get_crossreferences,
    get_reac_prop_with_ec,
    inchikey2pubchem_cid,
    load_cache_from_disk,
    metanetx2ec,
)
from cobramod.debug import debug_log
from cobramod.parsing.db_version import DataVersionConfigurator
from numpy import NaN

debug_log.setLevel(DEBUG)
data_conf = DataVersionConfigurator()
data_conf.force_same_version = True


class TestCrossReferences(TestCase):
    @patch("requests.get")
    def test_inchikey2pubchem_cid(self, mocked_post):
        with tempfile.TemporaryDirectory() as directory:
            directory = Path(directory)
            mocked_post.return_value.status_code = 200
            mocked_post.return_value.text = "154"

            value = inchikey2pubchem_cid("Test_ID", directory)
            self.assertEqual("154", value)

            # check that the local cache is used
            mocked_post.return_value.text = "155"
            value = inchikey2pubchem_cid("Test_ID", directory)
            self.assertEqual(value, "154")

            value = inchikey2pubchem_cid(["Test_ID2", "Test_ID3"], directory)
            self.assertEqual(["155", "155"], value)

    def test_load_cache_from_disk(self):
        with tempfile.TemporaryDirectory() as directory:
            directory = Path(directory)

            df = load_cache_from_disk("test", directory)
            result = pd.DataFrame({"ID": [], "XRefs": []})

            pd.testing.assert_frame_equal(result, df)

            dict = {"ID": "new_ID", "XRefs": "test"}

            # result = df.append(dict, ignore_index=True)
            result = pd.concat([df, pd.DataFrame([dict])], ignore_index=True)
            (directory / "XRef").mkdir(exist_ok=True)
            result.to_feather(directory / "XRef" / str("test" + ".feather"))
            load_cache_from_disk.cache_clear()
            df = load_cache_from_disk("test", directory)

            pd.testing.assert_frame_equal(result, df)

    @patch("requests.post")
    def test_get_crossreferences(self, mocked_post):
        with tempfile.TemporaryDirectory() as directory:
            directory = Path(directory)
            mocked_post.return_value.status_code = 200
            mocked_post.return_value.json.return_value = {
                "hmdb:HMDB62758": {
                    "InChI": "InChI=1S/C3H8O10P2/"
                    "c4-2(1-12-14(6,7)8)3(5)13-15(9,10)11/"
                    "h2,4H,1H2,(H2,6,7,8)(H2,9,10,11)",
                    "SMILES": "O=C(OP(=O)(O)O)C(O)COP(=O)(O)O",
                    "reference": "chebi:89363",
                    "name": "Glyceric acid 1,3-biphosphate",
                    "InChIkey": "LJQLQCAXBUHEAZ-UHFFFAOYSA-N",
                    "xrefs": [
                        "CHEBI:89363",
                        "chebi:89363",
                        "deprecated:MNXM733151",
                        "hmdb:HMDB0001270",
                        "hmdb:HMDB0062758",
                        "hmdb:HMDB01270",
                        "hmdb:HMDB62758",
                        "sabiork.compound:29",
                        "sabiorkM:29",
                    ],
                    "mnx_id": "MNXM1108074",
                }
            }

            value = get_crossreferences("chem", "hmdb:HMDB62758", directory)

            expected = {
                "inchikey:LJQLQCAXBUHEAZ-UHFFFAOYSA-N",
                "chebi:89363",
                "deprecated:MNXM733151",
                "CHEBI:89363",
                "sabiork.compound:29",
                "sabiorkM:29",
                "hmdb:HMDB62758",
                "hmdb:HMDB0062758",
                "hmdb:HMDB0001270",
                "inchi:InChI=1S/C3H8O10P2/"
                "c4-2(1-12-14(6,7)8)3(5)13-15(9,10)11/"
                "h2,4H,1H2,(H2,6,7,8)(H2,9,10,11)",
                "metanetx.chemical:MNXM1108074",
                "hmdb:HMDB01270",
            }

            self.assertEqual(expected, value)

            mocked_post.return_value.json.return_value = "something_else"
            value = get_crossreferences("chem", "hmdb:HMDB62758", directory)

            self.assertEqual(value, expected)

            value = get_crossreferences(
                "chem", ["hmdb:HMDB62758", "hmdb:HMDB62758"], directory
            )
            self.assertEqual(value, expected)

    @patch("pandas.read_csv")
    def test_metanetx2ec(self, mock):
        with tempfile.TemporaryDirectory() as directory:
            directory = Path(directory)
            mock.return_value = pd.DataFrame(
                data={
                    "ID": ["test1", "test2"],
                    "mnx_equation": ["test1", "test2"],
                    "reference": ["test1", "test2"],
                    "classifs": ["1.1.1.1", "test2"],
                    "is_balanced": ["test1", "test2"],
                    "is_transport": ["test1", "test2"],
                }
            )

            result = metanetx2ec("test1", directory)
            self.assertEqual("1.1.1.1", result)

            result = metanetx2ec(
                "test1", directory, include_metanetx_specific_ec=True
            )
            self.assertEqual("1.1.1.1", result)

            self.assertRaises(KeyError, metanetx2ec, "test2", directory)

            result = metanetx2ec(
                "test2", directory, include_metanetx_specific_ec=True
            )
            self.assertEqual("test2", result)

    @patch("pandas.read_csv")
    def test_get_reac_prop_with_ec(self, mock):
        with tempfile.TemporaryDirectory() as directory:
            directory = Path(directory)
            mock.return_value = pd.DataFrame(
                data={
                    "ID": ["test1", "test2"],
                    "mnx_equation": ["test1", "test2"],
                    "reference": ["test1", "test2"],
                    "classifs": [NaN, "test2"],
                    "is_balanced": ["test1", "test2"],
                    "is_transport": ["test1", "test2"],
                }
            )

            expected = pd.DataFrame(
                data={
                    "ID": ["test2"],
                    "mnx_equation": ["test2"],
                    "reference": ["test2"],
                    "classifs": ["test2"],
                    "is_balanced": ["test2"],
                    "is_transport": ["test2"],
                }
            )

            result = get_reac_prop_with_ec(directory)
            pd.testing.assert_frame_equal(expected, result)

            mock.return_value = "something_else"
            result = get_reac_prop_with_ec(directory)
            pd.testing.assert_frame_equal(expected, result)

    def test_add2dict_unique(self):
        dictonary = {"A": "A", "B": "B"}

        add2dict_unique("A", ["A", "B"], dictonary)
        expected = {"A": ["A", "B"], "B": "B"}

        # check keys
        self.assertCountEqual(expected, dictonary)

        # check values
        for key, value in expected.items():
            self.assertCountEqual(value, dictonary[key])

        add2dict_unique("C", "A", dictonary)
        expected = {"A": ["A", "B"], "B": "B", "C": "A"}

        # check keys
        self.assertCountEqual(expected, dictonary)

        # chack values
        for key, value in expected.items():
            self.assertCountEqual(value, dictonary[key])

    @patch("requests.get")
    @patch("pandas.read_csv")
    @patch("requests.post")
    def test_add_crossreferences(self, mocked_post, mock_pandas, mock_get):
        metabolite = Metabolite()
        metabolite.annotation = {"hmdb": "HMDB62758"}
        reaction = Reaction()
        reaction.annotation = {"hmdb": "HMDB62758"}

        mocked_post.return_value.status_code = 200
        mocked_post.return_value.json.return_value = {
            "hmdb:HMDB62758": {
                "InChI": "InChI=1S/C3H8O10P2/"
                "c4-2(1-12-14(6,7)8)3(5)13-15(9,10)11/"
                "h2,4H,1H2,(H2,6,7,8)(H2,9,10,11)",
                "SMILES": "O=C(OP(=O)(O)O)C(O)COP(=O)(O)O",
                "reference": "chebi:89363",
                "name": "Glyceric acid 1,3-biphosphate",
                "InChIkey": "LJQLQCAXBUHEAZ-UHFFFAOYSA-N",
                "xrefs": [
                    "CHEBI:89363",
                    "chebi:89363",
                    "deprecated:MNXM733151",
                    "hmdb:HMDB0001270",
                    "hmdb:HMDB0062758",
                    "hmdb:HMDB01270",
                    "hmdb:HMDB62758",
                    "sabiork.compound:29",
                    "sabiorkM:29",
                ],
                "mnx_id": "MNXM1108074",
            }
        }

        mock_pandas.return_value = pd.DataFrame(
            data={
                "ID": ["MNXM1108074", "test2"],
                "mnx_equation": ["test1", "test2"],
                "reference": ["test1", "test2"],
                "classifs": ["1.1.1.1", "test2"],
                "is_balanced": ["test1", "test2"],
                "is_transport": ["test1", "test2"],
            }
        )

        mock_get.return_value.text = "154"

        with tempfile.TemporaryDirectory() as directory:
            directory = Path(directory)

            expected_annotations = {
                "hmdb": [
                    "HMDB01270",
                    "HMDB0001270",
                    "HMDB0062758",
                    "HMDB62758",
                ],
                "metanetx.chemical": "MNXM1108074",
                "sabiorkm": "29",
                "inchi": "InChI=1S/C3H8O10P2/"
                "c4-2(1-12-14(6,7)8)3(5)13-15(9,10)11/"
                "h2,4H,1H2,(H2,6,7,8)(H2,9,10,11)",
                "inchikey": "LJQLQCAXBUHEAZ-UHFFFAOYSA-N",
                "sabiork.compound": "29",
                "chebi": "CHEBI:89363",
                "pubchem.compound": "154",
            }

            add_crossreferences(metabolite, directory)
            # check keys
            self.assertCountEqual(expected_annotations, metabolite.annotation)

            for key, value in expected_annotations.items():
                self.assertCountEqual(value, metabolite.annotation[key])

            add_crossreferences(reaction, directory)
            expected_annotations = {
                "hmdb": [
                    "HMDB0062758",
                    "HMDB62758",
                    "HMDB01270",
                    "HMDB0001270",
                ],
                "metanetx.reaction": "MNXM1108074",
                "sabiorkm": "29",
                "chebi": "CHEBI:89363",
                "metanetx.chemical": "MNXM1108074",
                "sabiork.compound": "29",
                "inchi": "InChI=1S/C3H8O10P2/"
                "c4-2(1-12-14(6,7)8)3(5)13-15(9,10)11/"
                "h2,4H,1H2,(H2,6,7,8)(H2,9,10,11)",
                "inchikey": "LJQLQCAXBUHEAZ-UHFFFAOYSA-N",
                "pubchem.compound": "154",
                "ec-code": "1.1.1.1",
                "brenda": "1.1.1.1",
            }

            # check keys
            self.assertCountEqual(expected_annotations, reaction.annotation)

            for key, value in expected_annotations.items():
                self.assertCountEqual(value, reaction.annotation[key])


if __name__ == "__main__":
    print(f"CobraMod version: {cmod_version}")
    print(f"COBRApy version: {cobra_version}")

    unittest.main(verbosity=2, failfast=True)



"""Test variables

This module creates the textbook models for testing. These are based on the
original model "e_coli_core" from BiGG database, which are also included in
COBRApy.
"""
import pkg_resources
from cobra.io import read_sbml_model

textbook = read_sbml_model(
    pkg_resources.resource_filename("cobra", "data/textbook.xml.gz")
)

textbook_biocyc = read_sbml_model(
    pkg_resources.resource_filename("cobramod", "data/textbook_biocyc.sbml")
)
textbook_kegg = read_sbml_model(
    pkg_resources.resource_filename("cobramod", "data/textbook_kegg.sbml")
)



"""Sub-package parsing

This sub-package takes care of gathering, storing and parsing the data from
multiple servers

- bigg: Parsing of BiGG information
- biocyc: Parsing and retrieval of BioCyc families
- db_version: Database versioning configurator
- kegg: Parsing of KEGG information
- plantcyc: Gene parsing for Plantcyc
- solcyc: Gene parsing for SolCyc
"""



"""Data parsing for KEGG

This module handles the retrieval of data from KEGG into a local directory.
The possible type of data that can be downloaded:

- Metabolite: Identifiers that start with the letter C, e.g C00001
- Reactions: Identifiers that start with the letter R, e.g R00001. The
gene information for reactions is also included if the specie is specified
- Module Pathways: Identifiers that start with the letter M, e.g M00001
"""
from __future__ import annotations

from contextlib import suppress
from itertools import chain
from pathlib import Path
from typing import Any, Generator, Optional

import requests

import cobramod.utils as cmod_utils
from cobramod.debug import debug_log
from cobramod.error import WrongParserError

KO_LINK = "http://rest.kegg.jp/link/ko/"


def parse_metabolite_attributes(
    data: dict[str, list[str]], entry: str
) -> dict[str, Any]:
    formula = data.get("FORMULA", ["X"]).pop()

    # NOTE: replace e.g (HPO3)n

    if formula.find("("):
        formula = formula.replace("(", "")
        formula = formula.replace(")", "")
        formula = formula.replace("n", "")

    attributes: dict[str, Any] = {
        "name": data.get("NAME", [entry]).pop(),
        "formula": formula,
        # NOTE: BiGG compounds does not show charges at a specific
        # ph-Value
        "charge": 0,
        "xref": build_references(data),
    }
    return attributes


def parse_reaction_attributes(
    data: dict[str, list[str]], entry: str, genome: str, gene_directory: Path
) -> dict[str, Any]:
    reaction_str = data.get("EQUATION", []).pop()

    if not reaction_str:
        raise AttributeError(f"There is no reaction found for '{entry}'")

    position = cmod_utils.get_arrow_position(reaction_str)

    if position == -1:
        raise AttributeError(
            f"Reaction string for {entry} did not include an arrow in the "
            "equation"
        )
    reactants = [
        item.rstrip().strip()
        for item in reaction_str[:position].rstrip().strip().split("+")
    ]
    products = [
        item.rstrip().strip()
        for item in reaction_str[position + 3 :].rstrip().strip().split("+")
    ]

    arrow = reaction_str[position : position + 3]

    reaction_str = str()

    for side in (reactants, products):
        prefix = "r"

        if side is reactants:
            prefix = "l"

        for i, item in enumerate(side):
            try:
                coefficient, identifier = [
                    item.strip().rstrip() for item in item.split(" ")
                ]
            except ValueError:
                # In case no coefficient is given, it must be 1
                coefficient = "1"
                identifier = item

            reaction_str += f"{coefficient} {prefix}_{identifier}"

            if i < len(side) - 1:
                reaction_str += " + "

            if side is reactants and i == len(side) - 1:
                reaction_str += f" {arrow} "

    attributes = {
        "name": data.get("NAME", [entry]).pop(),
        "equation": reaction_str,
        "xref": build_references(data),
        "replacements": {},
        "transport": cmod_utils.is_transport(reaction_str),
        "genes": parse_genes(
            identifier=entry,
            directory=gene_directory,
            genome=genome,
        ),
    }
    return attributes


def data_from_string(raw: str) -> dict[str, list[str]]:
    """
    Formats most of the keys for KEGG data and returns a dictionary.
    """
    lines = [line for line in raw.split(sep="\n")]
    data: dict[str, list[str]] = dict()

    for line in lines:
        segment = line.split(" ")

        if segment[0] != "///" and segment[0]:
            data[segment[0]] = []

    actual_key: str = ""

    for line in lines:
        if "///" in line:
            break

        # Create a list for each new key
        key = None
        for entry in data.keys():
            if entry in line:
                key = entry
                break

        if not key:
            data[actual_key].append(line.strip().rstrip())
        else:
            actual_key = key
            line = line[line.find(key) + len(key) :].strip().rstrip()
            data[actual_key] = [line]
    return data


def build_references(data_dict: dict[str, list[str]]) -> dict[str, str]:
    """
    Return a dictionary, where the keys are the names of cross-references
    and keys their identifiers. If nothing is found, it will return None
    """
    references = dict()

    for line in data_dict.get("DBLINKS", []):
        items = [item.rstrip().strip() for item in line.split(":")]
        references[items[0]] = items[1]

    return references


def ko_generator(identifier: str) -> Generator[str, None, None]:
    """
    Returns a list with the corresponding KO-entries for given identifier.
    Otherwise it will raise a HTTPError
    """
    url_text = f"http://rest.kegg.jp/link/ko/{identifier}"
    try:
        response = requests.get(url_text)
        response.raise_for_status()
        return (
            single
            for single in response.text.split()
            # This statement removes identifier from the split. e.g
            # [rn:R00001, ko:k00001]
            if identifier not in single
        )
    except requests.HTTPError:
        raise requests.HTTPError(
            f'No K-Identifiers could be retrieved for "{identifier}".'
        )


def parse_ko_to_genes(
    string: str, reaction: str, genome: Optional[str]
) -> list[str]:
    """
    Returns with a list with the corresponding genes for given genome. String
    is the raw text that include the gene information. If Abbreviation or no
    genome is present then, an empty list is returned.
    """
    # Retrive the KO-identifiers
    text = string.replace("\t", " ").splitlines()

    genes: dict[str, list] = dict()
    # Separate the lines and append the gene identifier to the corresponding
    # genome
    for line in text:
        _, single = line.split()
        specie, identifier = single.split(":")
        try:
            genes[specie].append(identifier)
        except KeyError:
            genes[specie] = [identifier]
    # Return the corresponding gene or empty list
    if genome:
        try:
            return genes[genome]
        except KeyError:
            msg = (
                f'Reaction "{reaction}" does not have a "{genome}" '
                + "abbreviation as a specie. No genes will be added."
            )
            debug_log.warning(msg=msg)
            return []
    msg = (
        f'Nothing was specified in argument "genome". Reaction "{reaction}"'
        " will not include genes. Please modify if necessary."
    )
    debug_log.warning(msg=msg)
    return []


def retrieve_kegg_genes(directory: Path, identifier: str):
    """
    Stores the genes for given reaction in given directory. Function will call
    a HTTPError if nothing is found.
    """
    # Ignore pathways and compounds
    if identifier.startswith("M") or identifier.startswith("C"):
        return

    directory = directory.joinpath("KEGG", "GENES")

    if not directory.exists():
        directory.mkdir()

    # Retrieval of the Gene information
    filename = directory.joinpath(f"{identifier}_genes.txt")

    if not filename.exists():
        try:
            string = "+".join(ko_generator(identifier))
            url_text = f"http://rest.kegg.jp/link/genes/{string}"
            response = requests.get(url_text)
            response.raise_for_status()

            with open(file=filename, mode="w") as file:
                file.write(response.text)

        except requests.ConnectionError:
            # NOTE: Using Generator otherwise, server breaks connection
            generator = ko_generator(identifier)
            while True:
                try:
                    ko = next(generator)
                    url_text = f"http://rest.kegg.jp/link/genes/{ko}"
                    response = requests.get(url_text)
                    response.raise_for_status()

                    with open(file=filename, mode="a+") as file:
                        file.write(response.text)
                except requests.HTTPError:
                    raise requests.HTTPError(
                        f'Gene information for "{identifier}" unavailable'
                    )
                except StopIteration:
                    break


def parse_genes(
    directory: Path, identifier: str, genome: str
) -> dict[str, Any]:
    """
    From given KEGG dictionary returns a dictionary with the key
    "genes" which include a dictionary with the identifier and name of the
    gene; and the key "rule" for the COBRApy representation of the
    gene-reaction rule
    """
    genes: dict[str, str] = dict()
    rule = str()

    filename = directory.joinpath(f"{identifier}_genes.txt")
    with suppress(FileNotFoundError):
        with open(file=filename, mode="r") as file:
            genes_list = parse_ko_to_genes(
                string=str(file.read()), reaction=identifier, genome=genome
            )
            for gene in genes_list:
                genes[gene] = ""

            rule = " or ".join(genes.keys())

    if genes:
        rule = " or ".join(genes.keys())
        debug_log.warning(
            f'Gene-reaction rule for reaction "{identifier}" from KEGG '
            'set to "OR".'
        )
    return {"genes": genes, "rule": rule}


def get_graph(kegg_dict: dict) -> dict:
    """
    Returns dictionary with sequences for a graph, where the key the prior
    reaction is and the value, its successor; and a set with all participant
    reaction (vertex)
    """
    # Obtain sequences as a set and regular reactions
    sequences = [item.split(sep=" ")[0] for item in kegg_dict["REACTION"]]
    whole_set = set(
        chain.from_iterable(line.split(sep=",") for line in sequences)
    )
    reactions = [line.split(sep=",") for line in sequences]
    graph: dict[str, Any] = dict()

    line: list
    for index_j, line in enumerate(iterable=reactions):
        reaction: str
        for reaction in line:
            # If index not found, then ignore it
            with suppress(IndexError):
                # In KEGG the order of the reactions is defined by its order in
                # the sequence.
                parent = reaction
                children = reactions[index_j + 1]
                if len(children) == 1:
                    # Take as a string
                    children = children[0]
                else:
                    # All elements as a tuple
                    children = tuple(children)
                # Expand keys
                try:
                    # If at least 2
                    if isinstance(graph[parent], tuple):
                        graph[parent] += children
                        continue
                    # Else transform single string to tuple
                    graph[parent] = (graph[parent], children)
                except KeyError:
                    # Create new one
                    graph[parent] = children
    # Check that the graph includes all reactions
    for reaction in whole_set:
        try:
            graph[reaction]
        except KeyError:
            graph[reaction] = None
    if not graph:
        raise WrongParserError("Given root does not belong to a Pathway")
    return graph


def parse_pathway_attributes(
    data: dict[str, list[str]], entry: str
) -> dict[str, Any]:
    graph = get_graph(data)
    name = data.get("NAME", [entry])

    attributes = {
        "name": name.pop(),
        "xref": build_references(data),
        "pathway": graph,
    }
    return attributes



#!/usr/bin/env python3
"""Unitest for build of summary

This modules checks functions responsable for the creation of txt, csv and xlsx
files.
"""
import tempfile
from pathlib import Path
from unittest import TestCase, main

import numpy
import pandas
from cobra import Model
from cobramod.core.summary import DataModel, summary
from cobramod.debug import change_to_debug
from cobramod.test import textbook

change_to_debug()


class TestSummary(TestCase):
    def test_from_model(self):
        # Case 1 example cobra model
        data_model = DataModel.from_model(textbook)
        self.assertIsInstance(obj=data_model, cls=DataModel)

        # Case 2 Empty cobra model
        model = Model("0")
        data_model = DataModel.from_model(model=model)
        self.assertIsInstance(obj=data_model, cls=DataModel)

        for variable in vars(data_model).values():
            self.assertEqual(variable, [])

    def test_diff(self):
        # Preparation
        data_model = DataModel.from_model(textbook)
        data_model_empty = DataModel(
            {
                "reactions": [],
                "metabolites": [],
                "demands": [],
                "exchanges": [],
                "genes": [],
                "groups": [],
                "sinks": [],
            }
        )
        tiny_base = DataModel(
            {
                "reactions": ["ACALD", "ACONTa"],
                "metabolites": ["13dpg_c", "acald_c"],
                "demands": ["h2o_e", "lac__D_e"],
                "exchanges": ["EX_glc__D_e", "EX_lac__D_e"],
                "genes": ["b1241", "b0474"],
                "groups": ["PWY-1187"],
                "sinks": ["SK_GLUTATHIONE_c", "SK_GLY_c"],
            }
        )

        tiny_changed = DataModel(
            {
                "reactions": ["ACALD", "ACKr"],
                "metabolites": ["acald_c", "2pg_c"],
                "demands": ["h2o_e", "h_e"],
                "exchanges": ["EX_lac__D_e", "EX_h2o_e"],
                "genes": ["b1241", "s0001"],
                "groups": ["PWY-1187-b"],
                "sinks": ["SK_GLUTATHIONE_c", "SK_AMMONIUM_c"],
            }
        )

        tiny_changes = DataModel(
            {
                "reactions": ["ACONTa", "ACKr"],
                "metabolites": ["13dpg_c", "2pg_c"],
                "demands": ["lac__D_e", "h_e"],
                "exchanges": ["EX_glc__D_e", "EX_h2o_e"],
                "genes": ["b0474", "s0001"],
                "groups": ["PWY-1187", "PWY-1187-b"],
                "sinks": ["SK_GLY_c", "SK_AMMONIUM_c"],
            }
        )

        # CASE 1: no differences
        diff = data_model.diff(data_model)

        for variable in vars(diff).values():
            self.assertEqual(variable, [])

        # Case 2: compare with empty models
        diff_left = data_model.diff(data_model_empty)
        diff_right = data_model_empty.diff(data_model)
        dict_left = vars(diff_left)
        dict_right = vars(diff_right)

        for key, value in vars(data_model).items():
            self.assertEqual(value, dict_left[key])
            self.assertEqual(value, dict_right[key])

        # Case 3: check with tiny_base for unidirectional similarity
        diff_left = data_model.diff(tiny_base)
        diff_right = tiny_base.diff(data_model)
        dict_left = vars(diff_left)
        dict_right = vars(diff_right)

        for key, value in vars(data_model).items():
            self.assertNotEqual(value, dict_left[key])
            self.assertNotEqual(value, dict_right[key])
            self.assertEqual(dict_left[key], dict_right[key])

        # Case 4: check correct difference identification
        diff_left = tiny_base.diff(tiny_changed)
        diff_right = tiny_changed.diff(tiny_base)
        dict_left = vars(diff_left)
        dict_right = vars(diff_right)

        for key, value in vars(tiny_changes).items():
            self.assertCountEqual(value, dict_left[key])
            self.assertCountEqual(value, dict_right[key])

    def test_to_dataframe(self):
        # Preparation
        tiny_base = DataModel(
            {
                "reactions": ["ACALD", "ACONTa"],
                "metabolites": ["13dpg_c", "acald_c"],
                "demands": ["h2o_e", "lac__D_e"],
                "exchanges": ["EX_glc__D_e", "EX_lac__D_e"],
                "genes": ["b1241", "b0474"],
                "groups": ["PWY-1187"],
                "sinks": ["SK_GLUTATHIONE_c", "SK_GLY_c"],
            }
        )

        expected_dataframe = pandas.DataFrame(
            {
                "Model identifier": ["e_coli_core", numpy.nan],
                "Model name": ["None", numpy.nan],
                "Reactions": ["ACALD", "ACONTa"],
                "Exchange": ["EX_glc__D_e", "EX_lac__D_e"],
                "Demand": ["h2o_e", "lac__D_e"],
                "Sinks": ["SK_GLUTATHIONE_c", "SK_GLY_c"],
                "Metabolites": ["13dpg_c", "acald_c"],
                "Genes": ["b1241", "b0474"],
                "Groups": ["PWY-1187", numpy.nan],
                "New in Reactions": ["ACALD", "ACONTa"],
                "New in Exchange": ["EX_glc__D_e", "EX_lac__D_e"],
                "New in Demand": ["h2o_e", "lac__D_e"],
                "New in Sinks": ["SK_GLUTATHIONE_c", "SK_GLY_c"],
                "New in Metabolites": ["13dpg_c", "acald_c"],
                "New in Genes": ["b1241", "b0474"],
                "New in Groups": ["PWY-1187", numpy.nan],
            },
            dtype=pandas.StringDtype(),
        )

        dataframe = tiny_base._to_dataframe(textbook, tiny_base)

        # Case 1: check that the dataframe contains the correct values

        pandas.testing.assert_frame_equal(expected_dataframe, dataframe)

        # Case 2: same as before but now with deletions

        expected_dataframe = pandas.DataFrame(
            {
                "Model identifier": ["e_coli_core", numpy.nan],
                "Model name": ["None", numpy.nan],
                "Reactions": ["ACALD", "ACONTa"],
                "Exchange": ["EX_glc__D_e", "EX_lac__D_e"],
                "Demand": ["h2o_e", "lac__D_e"],
                "Sinks": ["SK_GLUTATHIONE_c", "SK_GLY_c"],
                "Metabolites": ["13dpg_c", "acald_c"],
                "Genes": ["b1241", "b0474"],
                "Groups": ["PWY-1187", numpy.nan],
                "New in Reactions": ["ACALD", "ACONTa"],
                "New in Exchange": ["EX_glc__D_e", "EX_lac__D_e"],
                "New in Demand": ["h2o_e", "lac__D_e"],
                "New in Sinks": ["SK_GLUTATHIONE_c", "SK_GLY_c"],
                "New in Metabolites": ["13dpg_c", "acald_c"],
                "New in Genes": ["b1241", "b0474"],
                "New in Groups": ["PWY-1187", numpy.nan],
                "Removed in Reactions": ["ACALD", "ACONTa"],
                "Removed in Exchange": ["EX_glc__D_e", "EX_lac__D_e"],
                "Removed in Demand": ["h2o_e", "lac__D_e"],
                "Removed in Sinks": ["SK_GLUTATHIONE_c", "SK_GLY_c"],
                "Removed in Metabolites": ["13dpg_c", "acald_c"],
                "Removed in Genes": ["b1241", "b0474"],
                "Removed in Groups": ["PWY-1187", numpy.nan],
            },
            dtype=pandas.StringDtype(),
        )

        dataframe = tiny_base._to_dataframe(textbook, tiny_base, tiny_base)

        pandas.testing.assert_frame_equal(expected_dataframe, dataframe)

    def test_to_excl(self):
        tiny_base = DataModel(
            {
                "reactions": ["ACALD", "ACONTa"],
                "metabolites": ["13dpg_c", "acald_c"],
                "demands": ["h2o_e", "lac__D_e"],
                "exchanges": ["EX_glc__D_e", "EX_lac__D_e"],
                "genes": ["b1241", "b0474"],
                "groups": ["PWY-1187"],
                "sinks": ["SK_GLUTATHIONE_c", "SK_GLY_c"],
            }
        )

        with tempfile.TemporaryDirectory() as directory:
            filename = Path(directory) / "summary.xlsx"
            tiny_base.to_excl(filename, textbook, tiny_base, tiny_base)

            self.assertTrue(expr=filename.exists())

    def test_to_csv(self):
        tiny_base = DataModel(
            {
                "reactions": ["ACALD", "ACONTa"],
                "metabolites": ["13dpg_c", "acald_c"],
                "demands": ["h2o_e", "lac__D_e"],
                "exchanges": ["EX_glc__D_e", "EX_lac__D_e"],
                "genes": ["b1241", "b0474"],
                "groups": ["PWY-1187"],
                "sinks": ["SK_GLUTATHIONE_c", "SK_GLY_c"],
            }
        )

        with tempfile.TemporaryDirectory() as directory:
            filename = Path(directory) / "summary.csv"
            tiny_base.to_csv(filename, textbook, tiny_base, tiny_base)

            self.assertTrue(expr=filename.exists())

    def test_to_txt(self):
        tiny_base = DataModel(
            {
                "reactions": ["ACALD", "ACONTa"],
                "metabolites": ["13dpg_c", "acald_c"],
                "demands": ["h2o_e", "lac__D_e"],
                "exchanges": ["EX_glc__D_e", "EX_lac__D_e"],
                "genes": ["b1241", "b0474"],
                "groups": ["PWY-1187"],
                "sinks": ["SK_GLUTATHIONE_c", "SK_GLY_c"],
            }
        )

        with tempfile.TemporaryDirectory() as directory:
            filename = Path(directory) / "summary.txt"
            tiny_base.to_txt(filename, textbook, tiny_base, tiny_base)

            self.assertTrue(expr=filename.exists())

    def test_summary(self):
        # Preparation
        test_model = textbook.copy()
        old_values = DataModel.from_model(textbook)

        # Case 1: None as filename
        with tempfile.TemporaryDirectory() as directory:
            directory = Path(directory)
            filename = None
            summary(test_model, old_values, filename=filename)

            for _ in directory.iterdir():
                self.fail("Summary created a file although None was passed")

        # Case 2: Excel as format
        with tempfile.TemporaryDirectory() as directory:
            directory = Path(directory)
            filename = directory / "summary.xlsx"
            summary(test_model, old_values, filename=filename)

            self.assertTrue(expr=filename.exists())

            for file in directory.iterdir():
                if file != filename:
                    self.fail("Summary created additional files")

        # Case 3: csv as format
        with tempfile.TemporaryDirectory() as directory:
            directory = Path(directory)
            filename = directory / "summary.csv"
            summary(test_model, old_values, filename=filename)

            self.assertTrue(expr=filename.exists())

            for file in directory.iterdir():
                if file != filename:
                    self.fail("Summary created additional files")

        # Case 4: txt as format
        with tempfile.TemporaryDirectory() as directory:
            directory = Path(directory)
            filename = directory / "summary.txt"
            summary(test_model, old_values, filename=filename)

            self.assertTrue(expr=filename.exists())

            for file in directory.iterdir():
                if file != filename:
                    self.fail("Summary created additional files")

        # Case 5: unknown format
        with tempfile.TemporaryDirectory() as directory:
            directory = Path(directory)
            filename = directory / "summary.unknown"
            summary(test_model, old_values, filename=filename)

            for _ in directory.iterdir():
                self.fail("Summary created a file!")


if __name__ == "__main__":
    main(verbosity=2)



"""Data parsing for Biocyc

This module handles the retrieval of data from BioCyc into a local directory.
The possible type of data that can be downloaded:

- Metabolites: Normally have an abbreviation or short name.
- Reactions: Can have the words "RXN" in the identifier. Enzymes can sometimes
be used instead. The gene information for the reactions is included if found.
"""
from __future__ import annotations

import urllib.parse
import xml.etree.ElementTree as et
from contextlib import suppress
from pathlib import Path
from typing import Any

import requests

import cobramod.error as cmod_error
import cobramod.utils as cmod_utils
from cobramod.debug import debug_log


def build_cross_references_xml(root: Any) -> dict[str, str]:
    """
    Returns a dictionary with the corresponding cross-references and their
    identifiers.
    """
    references = dict()
    try:
        for child in root.findall("*/dblink"):
            references[child.find("dblink-db").text] = child.find(
                "dblink-oid"
            ).text
        return references
    except IndexError:
        raise IndexError("No references can be found in given XML.")


def reaction_str_from_xml(root: et.Element) -> str:
    reactants = root.findall("./Reaction/left")
    products = root.findall("./Reaction/right")

    if not reactants or not products:
        raise AttributeError(
            "The current et.Element does not show any properties of being a"
            "reaction."
        )
    reaction_str: str = ""

    for side in (reactants, products):
        prefix = "r"

        if side is reactants:
            prefix = "l"

        for i, item in enumerate(side):
            element = item.find("coefficient")
            if not element:
                coefficient: float = 1.0

            else:
                value = element.text
                if not value:
                    raise AttributeError("Given Elements does not include text")

                coefficient = float(value)

            element = item.find("*/[@frameid]")

            if element is None:
                raise AttributeError("Metabolite was not found in the XML root")
            else:
                identifier = element.attrib["frameid"].strip().rstrip()

            reaction_str += f"{coefficient} {prefix}_{identifier}"

            if i < len(side) - 1:
                reaction_str += " + "

            if side is reactants and i == len(side) - 1:
                reaction_str += f" {get_direction_from_xml(root)} "

    return reaction_str


def get_direction_from_xml(root: et.Element) -> str:
    """
    Returns the direction of the chemical reaction
    """
    element = root.find("*/reaction-direction")
    if element is None:
        raise AttributeError(
            "Attribute 'reaction-direction' was not found in the given Element"
        )
    text = element.text
    if not text:
        raise AttributeError(
            "Given element does not include text and thus, reversibility "
            "cannot be infered"
        )

    reversibility = {
        "REVERSIBLE": "<->",
        "RIGHT-TO-LEFT": "<--",
        "PHYSIOL-RIGHT-TO-LEFT": "<--",
        "LEFT-TO-RIGHT": "-->",
        "PHYSIOL-LEFT-TO-RIGHT": "-->",
    }
    return reversibility[text]


def parse_reaction_attributes(
    root: et.Element, entry: str, gene_directory: Path
) -> dict[str, Any]:
    name = root.find("*[@ID]/enzymatic-reaction/*/common-name")
    if name is not None and name.text:
        name = name.text
    else:
        name = entry
    equation = reaction_str_from_xml(root)

    attributes = {
        "name": name,
        "equation": equation,
        "xref": build_cross_references_xml(root),
        "replacements": {},
        "transport": cmod_utils.is_transport(equation),
        "genes": parse_genes(entry, gene_directory),
    }
    return attributes


def parse_metabolite_attributes(root: et.Element, entry: str):
    try:
        element = root.find("./*/cml/*")

        if element is None:
            raise AttributeError("No compound was found in given XML root")
        name = element.attrib["title"]

    except AttributeError:
        name = entry

    try:
        element = root.find("./*/cml/*/formula")
        if element is None:
            raise AttributeError("No formula was found in given XML root")
        formula = element.attrib["concise"].replace(" ", "")

        element = root.find("./*/cml/molecule")
        if element is None:
            raise AttributeError("No formula was found in given XML root")
        charge = float(element.attrib["formalCharge"])

    # Must be Protein
    except AttributeError:
        formula = "X"
        charge = 0

    attributes: dict[str, Any] = {
        "name": name,
        "formula": formula,
        "charge": charge,
        "xref": build_cross_references_xml(root),
    }
    return attributes


def get_graph(root: et.Element) -> dict:
    """
    Return a directed graph from given XML Element. A key represent the
    parent reaction and the value is the child. A parent can have multiple
    children as tuples. End-reactions have a NoneType as child.

    Args:
        root (Element): An XML element, which represent a XML file with the
            information of a pathway.

    Returns:
        Dict: Directed graph from root.

    Raises:
        WrongParserError: If given root does not represent a pathway.
    """
    # Verify non-superpathway
    element = root.find("*[@frameid]")
    if element is None:
        raise TypeError("Not a valid Element")

    identifier = element.get("frameid")

    for pathway in element.findall("*/Pathway"):
        frameid = pathway.get("frameid")

        if frameid is None:
            raise TypeError("Not a valid Element")

        if "Super" in frameid:
            msg = (
                f'Pathway "{identifier}" was identified as a '
                + "superpathway. This type of pathway does not normally "
                + "included all reactions. Please add the corresponding "
                + "sub-pathways singlely!. "
                # NOTE: ToDo
                + "For the moment, CobraMod does not support them"
            )
            debug_log.error(msg=msg)
            raise cmod_error.SuperpathwayException(msg)

    # identifiers = [
    #     a.attrib["frameid"]
    #     for a in root.findall("*/reaction-ordering/Reaction")
    # ]
    graph: dict[str, Any] = dict()
    reactions = set()

    for line in root.findall("*/reaction-ordering"):
        # Get reactions and add them to set
        element = line.find("*/[@frameid]")
        if element is None:
            raise TypeError("Not a valid Element")
        child = element.attrib["frameid"]

        element = line.find("predecessor-reactions/")
        if element is None:
            raise TypeError("Not a valid Element")
        parent = element.attrib["frameid"]

        reactions.add(parent)
        reactions.add(child)
        # Expand keys
        try:
            # If at least 2
            if isinstance(graph[parent], tuple):
                graph[parent] = graph[parent] + (child,)
                continue
            # Else transform single to tuple
            graph[parent] = (graph[parent], child)
        except KeyError:
            # Create new one
            graph[parent] = child

    # Find missing elements and insert a None
    for reaction in reactions:
        try:
            graph[reaction]
        except KeyError:
            graph[reaction] = None

    if not graph:
        # It could be a single-reaction pathway. Give a None since it is just
        # one
        try:
            element = root.find("*/reaction-layout/Reaction")
            if element is None:
                raise TypeError("Not a valid Element")
            name = element.attrib["frameid"]
            graph[name] = None
        except AttributeError:
            raise cmod_error.WrongParserError(
                "Given root does not belong to a Pathway"
            )

    return graph


def parse_pathway_attributes(root: et.Element, entry: str) -> dict[str, Any]:
    graph = get_graph(root)
    element = root.find("*[@frameid]")

    if element is not None:
        name = element.attrib["frameid"]
    else:
        name = entry

    attributes = {
        "name": name,
        "xref": build_cross_references_xml(root),
        "pathway": graph,
    }
    return attributes


def retrieve_gene_information(directory: Path, identifier: str, database: str):
    """
    Retrieve the gene information for given reaction in a specific database
    Args:
        directory (Path): Directory to store and retrieve local data.
        identifier (str): original identifier
        database (str): Name of the database. Some options: "META", "ARA".
            Full list in: "https://biocyc.org/biocyc-pgdb-list.shtml"
    Raises:
        HTTPError: If given reaction does not have gene information available
    """
    # GENES directory will depend from the sub-database
    directory = directory.joinpath(database, "GENES")

    if not directory.exists():
        directory.mkdir()

    # Retrieval of the Gene information
    filename = directory.joinpath(f"{identifier}_genes.xml")
    if database == "META":
        msg = (
            f'Object "{identifier}" retrieved from "META". Please use '
            "a sub-database to add proper genes. "
            "Skipping retrieval of gene information."
        )
        debug_log.warning(msg)

        return

    if not filename.exists():
        # This URL will not necessarily raise exception
        encoded_id = urllib.parse.quote(identifier, safe="")

        url_text = (
            f"https://websvc.biocyc.org/apixml?fn=genes-of-reaction&id="
            f"{database}:{encoded_id}&detail=full"
        )

        # FIXME: find a better way
        user, pwd = cmod_utils.get_credentials(
            Path.cwd().joinpath("credentials.txt")
        )
        s = requests.Session()
        s.post(
            "https://websvc.biocyc.org/credentials/login/",
            data={"email": user, "password": pwd},
        )
        response = s.get(url_text)
        s.close()
        try:
            response.raise_for_status()
            root = et.fromstring(response.text)
            # Check for results in root
            tree = et.ElementTree(root)

            element = tree.find("*/num_results")

            if not isinstance(element, et.Element):
                raise AttributeError("No gene information available")

            # Ignore metabolites
            if element.text == "0":
                return

            tree.write(str(filename))
            debug_log.info(
                f'Object "{identifier}_gene.xml" saved in '
                f'directory "{database}/GENES".'
            )
        except requests.HTTPError:
            raise AttributeError(
                f"Object {identifier} does not have gene information"
            )


def parse_genes(identifier: str, directory: Path) -> dict:
    """
    Returns a dictionary with the corresponding genes and gene-reaction rules.
    This function will try to read a file for given identifier. If nothing is
    found, the dictionary will have empty entries.
    """
    rule = str()
    genes = dict()
    # Get the information and check if Genes can be found to be parsed
    with suppress(FileNotFoundError):
        tree = et.parse(directory.joinpath(f"{identifier}_genes.xml")).getroot()

        if not isinstance(tree, et.Element):
            raise TypeError("Given root is not a valid Element object")

        for gene in tree.findall("Gene"):
            name = gene.attrib.get("frameid")

            if name is None:
                raise TypeError("Given root is not a valid Element object")

            genes[name] = name

        # Assuming rule
        rule = " or ".join(genes.keys())
        debug_log.warning(
            f'Gene-reaction rule for reaction "{identifier}" set'
            + ' to "OR". Please modify it if necessary.'
        )
    return {"genes": genes, "rule": rule}



"""Gene parsing for SolanaCyc

The main functions to parse Metabolites and Reactions are found in the module
Biocyc
"""
import urllib.parse
import xml.etree.ElementTree as et
from pathlib import Path
from typing import Any

import requests

from cobramod.debug import debug_log


def retrieve_gene_information(directory: Path, identifier: str, database: str):
    """
    Retrieve the gene information for given reaction in a specific database

    Args:
        directory (Path): Directory to store and retrieve local data.
        identifier (str): original identifier
        database (str): Name of the database. Some options: "META", "LYCO".
            Full list in: "https://solcyc.sgn.cornell.edu/xmlquery?dbs"

    Raises:
        HTTPError: If given reaction does not have gene information available
    """
    # GENES directory will depend from the sub-database
    directory = directory.joinpath(database, "GENES")

    if not directory.exists():
        directory.mkdir()

    # Retrieval of the Gene information
    filename = directory.joinpath(f"{identifier}_genes.xml")
    if database == "META":
        msg = (
            f'Object "{identifier}" retrieved from "META". Please use '
            "a sub-database to add proper genes. "
            "Skipping retrieval of gene information."
        )
        debug_log.warning(msg)

        return

    if not filename.exists():
        # This URL will not necessarily raise exception
        encoded_id = urllib.parse.quote(identifier, safe="")

        url_text = (
            f"https://solcyc.sgn.cornell.edu/apixml?fn=genes-of-reaction&id="
            f"{database}:{encoded_id}&detail=full"
        )
        response = requests.get(url_text)
        try:
            response.raise_for_status()
            root = et.fromstring(response.text)
            # Check for results in root
            tree: Any = et.ElementTree(root)

            element = tree.find("*/num_results")
            if not isinstance(element, et.Element):
                raise AttributeError("No gene information available")

            tree.write(str(filename))
            debug_log.info(
                f'Object "{identifier}_gene.xml" saved in '
                f'directory "{database}/GENES".'
            )
        except requests.HTTPError:
            raise AttributeError(
                f"Object {identifier} does not have gene information"
            )



"""Errors for CobraMod

This module creates special errors for CobraMod. Read each error for their
explanation.
"""
from cobramod.debug import debug_log


class UserInterruption(Exception):
    """
    Exception that is used if the input of a user interrupts the program flow.
    """

    pass


class NoDelimiter(Exception):
    """
    Simple Error that should be raised when a string does not include the
    delimiter "|"
    """

    def __init__(self, string: str):
        """
        Args:
            string (str): String with the format problem
        """
        msg = f'No delimiter "|" was found in \n{string}'
        super().__init__(msg)


class GraphKeyError(Exception):
    """
    Simple Error that should be raised when a value is missing as key in a
    graph
    """

    pass


class FoundInPairError(Exception):
    """
    Simple Error that is raised when the pair of a PairDictionary has already
    given key name.
    """

    pass


class NodeAttributeError(Exception):
    """
    Simple Error that should be raised when a
    :func:`cobramod.visualization.Node` cannot identify its type.
    """

    pass


class WrongParserError(Exception):
    """
    Simple Error that should be raised if a method cannot handle the parsing.
    """

    pass


class WrongDataError(Exception):
    """
    Simple Error that should be raised if COBRApy object is not the correct
    one.
    """

    pass


class WrongSyntax(Exception):
    """
    Simple Error that should be raised if a specific syntax is invalid. For
    instance, when strings do not have a certain number of items when splitted.
    """

    pass


class PatternNotFound(Exception):
    """
    Simple error that should be raised when a pattern is not found in a item.
    e.g. a substring.
    """

    pass


class NoIntersectFound(Exception):
    """
    Simple error that should be raised when no intersect is found between two
    sets.
    """

    pass


class NoGeneInformation(Exception):
    """
    Simple error that should be raised when given object has no gene
    information in the database.
    """

    pass


class SuperpathwayException(Exception):
    """
    Simple Warning that should be raised if a pathway is identified as a
    Superpathway
    """

    pass


class UnbalancedReaction(Exception):
    """
    Simple Error that should be raised if a reaction has wrong mass balance.
    """

    def __init__(self, identifier: str, dict_balance: str):
        """
        Args:
            reaction (str): identifier of the reaction.
        """
        msg = (
            f'Reaction "{identifier}" unbalanced. Following atoms are '
            + f"affected. Please verify: {dict_balance}"
        )
        debug_log.critical(msg)
        super().__init__(msg)



"""
Database versioning configurator

This modules includes the class DataVersionConfigurator which is in charge of
obtaining and comparing the version from the obtained metabolic data.
It uses the same structure of using a Singleton for the configuration in
COBRApy
"""
from pathlib import Path
from typing import Optional, Union

import pandas as pd
from cobra.core.singleton import Singleton

import cobramod.error as cmod_error
from cobramod.debug import debug_log


class DataVersionConfigurator(metaclass=Singleton):
    def __init__(self):
        self.ignore_db_versions: bool = False
        self.database_version: Optional[pd.DataFrame] = None
        self.force_same_version: bool = False

    def get_database_version(self, directory: Union[str, Path]) -> pd.DataFrame:
        """
        Loads and returns the database versioning file.

        Args:
            directory (Path): The folder used for storing data.

        Returns:
            (pd.DataFrame): A DataFrame containing the orgid and version, at
                the time of the first retrieval, for all databases used so far.
        """
        if isinstance(directory, str):
            directory = Path(directory)

        if self.database_version is not None:
            return self.database_version
        try:
            file = directory / "DatabaseVersions.csv"
            self.database_version = pd.read_csv(file)
        except FileNotFoundError:
            self.database_version = pd.DataFrame(columns=["orgid", "version"])

        return self.database_version

    def get_local_databases(self, directory: Union[Path, str]) -> pd.Series:
        if isinstance(directory, str):
            directory = Path(directory).absolute()
        databases = self.get_database_version(directory)
        databases = databases["orgid"]

        return databases

    def check_database_version(
        self, directory: Union[str, Path], database: str, version: Optional[str]
    ):
        """
        Function to compare the saved database version with the one
        of the retrieved data.

        Args:
            directory (Path): The folder used for storing data.
            database (str): Identifier of the database.
            version (str): The version of the database.
        """
        if isinstance(directory, str):
            directory = Path(directory).absolute()

        if self.database_version is None:
            self.get_database_version(directory=directory)

        assert isinstance(self.database_version, pd.DataFrame)
        row = self.database_version.loc[
            self.database_version["orgid"] == database
        ]

        if row.shape[0] == 0:
            return self.set_database_version(directory, database, version)

        if row.shape[0] > 1:
            exit()

        # NOTE: not all version use the same versioning
        expected_version = str(row["version"].tolist()[0])

        # row can not be more than one value but python does not know that
        # => any

        if expected_version != version:
            msg = (
                "Versions of {} do not match. Remote has version {} "
                "and local version is {}.".format(
                    database, version, expected_version
                )
            )
            debug_log.warning(msg)

            if self.ignore_db_versions:
                return

            if self.force_same_version:
                raise Exception(
                    "DB Configuration: force_same_version is True\n"
                    "Database differs from local dabatase and retrieved data!"
                )
            else:
                while True:
                    choice = input("Ignore version mismatch? (Y)es (N)o: ")

                    if choice.lower() == "y":
                        self.ignore_db_versions = True
                        break

                    if choice.lower() == "n":
                        raise cmod_error.UserInterruption(
                            "Interrupted by user input"
                        )

    def set_database_version(
        self, directory: Union[str, Path], database: str, version: Optional[str]
    ) -> bool:
        """
        Adds the version of a database to the local data versioning file.

        Args:
            directory (Path): The folder used for storing data.
            database (str): Identifier of the database.
            version (str): The version of the database.

        Returns:
             (bool): Returns True if the addition was successful.

        """
        if isinstance(directory, str):
            directory = Path(directory).absolute()

        if self.database_version is None:
            self.get_database_version(directory=directory)

        assert isinstance(self.database_version, pd.DataFrame)
        self.database_version = pd.concat(
            [
                self.database_version,
                pd.DataFrame({"orgid": [database], "version": [version]}),
            ],
            ignore_index=True,
        )

        self.database_version.to_csv(
            path_or_buf=str(directory / "DatabaseVersions.csv"), index=False
        )

        return True



"""
Retrieval of data from multiple databases

This module includes the class 'Data', that is responsable for the internal
data representation for this package. The class Data includes methods for
parsing itself into a Reaction, Metabolite or Pathway.

Additionally, this package includes high level functions such as `get_data`,
`file_to_Data_class` or `get_response` that handles the data retrieval or
convertion
"""
from __future__ import annotations

import json
import urllib.parse
import warnings
import xml.etree.ElementTree as et
from pathlib import Path
from typing import Any, Literal, Optional, Union

import cobra.core as cobra_core
import requests

import cobramod.utils as cmod_utils
from cobramod.core import creation, genes
from cobramod.debug import debug_log
from cobramod.parsing import bigg, biocyc, kegg, plantcyc, solcyc
from cobramod.parsing import db_version as cmod_db

db_configuration = cmod_db.DataVersionConfigurator()

BIOCYC = "https://websvc.biocyc.org/getxml?id="
PMN = "https://pmn.plantcyc.org/getxml?id="
KEGG = "https://rest.kegg.jp/get/"
# NOTE: deprecation for v2
SOLCYC = "https://solcyc.sgn.cornell.edu/getxml?id="

EXTENSIONS = [".xml", ".txt", ".json"]


class Databases:
    """
    Simple object that shows the information about the database
    """

    def __init__(self):
        self.msg = (
            "CobraMod supports BioCyc, the Plant Metabolic Network (PMN), "
            "KEGG and BiGG Models repository. "
            "BioCyc includes around 18.000 sub-databases. A complete list "
            "for BioCyc can be found at: "
            "'https://biocyc.org/biocyc-pgdb-list.shtml'.\n"
            "The database-specific identifiers can be found in the URL of the "
            "respective data."
        )

    def __repr__(self):
        """
        Returns a string about the definition of "identifier" and where
        to obtain them.
        """
        self.msg2 = (
            "CobraMod uses abbreviations to represent the databases or "
            "sub-databases:\n"
            "Abbreviation   Database Name\n"
            "META           Biocyc, subdatabase MetaCyc\n"
            "ARA            Biocyc, subdatabase AraCyc\n"
            "\n"
            "KEGG           Kyoto encyclopedia of Genes and Genomes\n"
            "BIGG           BiGG Models\n"
            "\n"
            "PMN:META       Plantcyc, subdatabase META\n"
            "PMN:ARA        Plantcyc, subdatabase ARA\n"
            "This applies for all subdatabases from SolCyc, BioCyc and Plantcyc\n"
        )

        return self.msg + self.msg2

    def _repr_html_(self):
        """
        Returns a HTML string about the definition of "identifier" and where
        to obtain them.
        """
        return """
<p>CobraMod supports BioCyc, the Plant Metabolic Network (PMN), KEGG and BiGG Models repository. BioCyc includes around 18.000 sub-databases. A complete list for BioCyc can be found at: 'https://biocyc.org/biocyc-pgdb-list.shtml'.<br
    />The database-specific identifiers can be found in the URL of the respective data.CobraMod uses abbreviations to represent the databases or sub-databases:
</p>
<table style="width: 50%; border-collapse: collapse;" border="1">
    <tbody>
        <tr>
            <td style="width:
        50%;"><strong>Abbreviation</strong></td>
            <td style="width:
            50%;"><strong>Database name</strong></td>
        </tr>
        <tr>
            <td style="width: 50%;">META</td>
            <td style="width: 50%;">Biocyc, subdatabase MetaCyc</td>
        </tr>
        <tr>
            <td style="width:
                50%;">ARA</td>
            <td style="width: 50%;">Biocyc, subdatabase AraCyc
            </td>
        </tr>
        <tr>
            <td style="width: 50%;">KEGG</td>
            <td style="width: 50%;">Kyoto encyclopedia of Genes and Genomes
            </td>
        </tr>
        <tr>
            <td style="width: 50%;">BIGG</td>
            <td style="width: 50%;">BiGG Models</td>
        </tr>
        <tr>
            <td style="width: 50%;">PMN:META</td>
            <td style="width:
                    50%;">PlantCyc, subdatabase META</td>
        </tr>
        <tr>
            <td style="width: 50%;">PMN:ARA</td>
            <td style="width:
                        50%;">PlantCyc, subdatabase ARA</td>
        </tr>
    </tbody>
</table>
<p>This applies for all subdatabases from BioCyc and Plantcyc
</p>
        """


available_databases = Databases()

# NOTE: bigg uses an extra function
databases = {
    "BIOCYC": BIOCYC,
    "PMN": PMN,
    "KEGG": KEGG,
    "BIGG": "",
    "SOL": SOLCYC,
}


class Data:
    identifier: str
    mode: str
    database: str
    attributes: dict[str, Any]
    path: Path
    model_id: str
    genome: str
    version: str
    """
    CobraMod's Data object. This class is responsable for converting
    information into COBRApy objects
    """

    def __init__(
        self,
        identifier: str,
        attributes: dict[str, str],
        mode: Literal["Pathway", "Reaction", "Metabolite"],
        database: str,
        path: Path,
        model_id: str,
        version: str,
    ):
        self.mode = mode
        self.identifier = identifier
        self.attributes = attributes
        self.database = database
        self.path = path
        self.model_id = model_id
        self.genome = ""
        self.version = version

    def __repr__(self):
        return f"Data [{self.mode}: {self.identifier}]"

    @classmethod
    def from_json(
        cls, entry: str, data: dict[str, Any], database: str, path: Path
    ):
        """
        Creates an instance from a JSON file. Database: BIGG
        """
        is_compound = data.get("formulae", data.get("formula", None))
        model_id = path.parent.name

        if path.parents[1].joinpath("database_version").exists():
            with open(path.parents[1].joinpath("database_version"), "r") as f:
                version = json.load(f).get("bigg_models_version")
        else:
            info_response = requests.get(
                "http://bigg.ucsd.edu/api/v2/database_version"
            )
            info_response.raise_for_status()

            with path.parents[1].joinpath("database_version").open(
                mode="w+"
            ) as f:
                f.write(info_response.text)
            version = info_response.json().get("bigg_models_version")

        if version is None:
            raise Exception("Version of the BIGG JSON cannot be empty!")

        # Non universal models includes suffices
        if model_id != "universal":
            entry = entry[: entry.find("_")]

        if is_compound:
            mode = "Metabolite"
            attributes = bigg.parse_metabolite_attributes(data)
        else:
            mode = "Reaction"
            attributes = bigg.parse_reaction_attributes(data, entry)

        return cls(entry, attributes, mode, database, path, model_id, version)

    @classmethod
    def from_text(
        cls, entry: str, text: str, database: str, path: Path, genome: str
    ):
        """
        Creates an instance from plain text. Database: KEGG
        """
        data: dict[str, list[str]] = kegg.data_from_string(text)
        mode = data["ENTRY"][0].split()[-1]
        gene_path = path.parent.joinpath("GENES")

        if path.parent.joinpath("database_version").exists():
            with open(path.parent.joinpath("database_version"), "r") as f:
                version = cmod_utils.kegg_info_to_version(f.read())
        else:
            info_response = requests.get("http://rest.kegg.jp/info/kegg")
            info_response.raise_for_status()

            with path.parent.joinpath("database_version").open(mode="w+") as f:
                f.write(info_response.text)
            version = cmod_utils.kegg_info_to_version(info_response.text)

        if version is None:
            raise Exception("Version of the BIGG JSON cannot be empty!")

        if mode == "Compound":
            mode = "Metabolite"
            attributes = kegg.parse_metabolite_attributes(data, entry)

        elif mode == "Reaction":
            attributes = kegg.parse_reaction_attributes(
                data, entry, genome, gene_path
            )

        elif mode == "Module":
            mode = "Pathway"
            attributes = kegg.parse_pathway_attributes(data, entry)
        else:
            raise AttributeError(f"Cannot parse type '{mode}'")

        obj = cls(entry, attributes, mode, database, path, "", version)
        obj.genome = genome
        return obj

    @classmethod
    def from_xml(cls, entry: str, root: et.Element, database: str, path: Path):
        """
        Creates an instance from an XML file. Database: Biocyc families
        """
        version = root.find("metadata/PGDB")
        if version is not None:
            version = version.attrib.get("version")

        if version is None:
            raise Exception("Version for the XML cannot be empty!")

        gene_path = path.parent.joinpath("GENES")

        if (
            root.findall("Compound")
            or root.findall("Protein")
            or root.findall("RNA")
        ):
            mode = "Metabolite"
            attributes = biocyc.parse_metabolite_attributes(root, entry)

        elif root.findall("Reaction"):
            mode = "Reaction"
            attributes = biocyc.parse_reaction_attributes(
                root, entry, gene_path
            )

        elif root.findall("Pathway"):
            mode = "Pathway"
            attributes = biocyc.parse_pathway_attributes(root, entry)
        else:
            raise AttributeError(
                "Cannot infere the object type from given et.Element"
            )
        return cls(entry, attributes, mode, database, path, "", version)

    def parse(
        self,
        model: cobra_core.Model,
        compartment: str,
        replacement: dict[str, str] = {},
    ) -> cobra_core.Object:
        """
        Parses information and creates a cobrapy Object. This can be a
        Metabolite or Reaction.

        Args:
            model (cobra.core.Model): Model to search for information
            compartment (str): Location of the object
            replacement (dict[str, str]): In case of Reactions, the dictionary
                represents the metabolites to replace. The key is the
                identifier of the object to replaced and the value is the new
                identifier to use instead. Defaults to an empty dictionary

        Returns:
            cobra.core.Object This can be a Reaction or Metabolites

        """
        identifier = replacement.get(self.identifier, self.identifier)
        identifier = f"{identifier.replace('-','_')}_{compartment}"

        if self.mode == "Metabolite":
            try:
                return model.metabolites.get_by_id(identifier)

            except KeyError:
                return cmod_utils.build_metabolite(
                    identifier,
                    self.attributes["formula"],
                    self.attributes["name"],
                    self.attributes["charge"],
                    compartment,
                )
        elif self.mode == "Reaction":
            xref = cmod_utils.find_intersection(
                dictlist=model.reactions,
                query=self.attributes["xref"],
                revert=True,
            )
            if xref:
                self.identifier = xref

            try:
                reaction = model.reactions.get_by_id(identifier)
                if not isinstance(reaction, cobra_core.Reaction):
                    raise TypeError("Not a valid COBRApy Reaction")
                genes.genes_to_reaction(reaction, self.attributes["genes"])
                return reaction

            except KeyError:
                reaction = cobra_core.Reaction(
                    identifier, self.attributes["name"]
                )
                model.add_reactions([reaction])

                reaction_str = ""
                part: str
                for part in self.attributes["equation"].split(" "):
                    # Removal of prefixes
                    if cmod_utils.is_compound(part):
                        part = part[2:]
                        part += f"_{compartment}"

                    part = replacement.get(part, part)

                    reaction_str += f"{part} "

                if self.attributes["transport"]:
                    reaction_str = cmod_utils.convert_to_transport(
                        reaction_str, compartment
                    )
                    msg = (
                        f'Reaction "{reaction.id}" has the same metabolite on'
                        " both sides of the equation (e.g transport reaction). "
                        "COBRApy ignores these metabolites. To avoid this, by "
                        "default, CobraMod will assign the reactant side to "
                        "the extracellular compartment. Please curate the "
                        "reaction if necessary."
                    )
                    debug_log.warning(msg=msg)

                directory = self.path.parents[1]

                if self.database == "BIGG":
                    directory = self.path.parents[2]

                build_reaction_from_str(
                    model,
                    reaction,
                    reaction_str,
                    directory,
                    self.database,
                    self.model_id,
                    replacement,
                )

                genes.genes_to_reaction(reaction, self.attributes["genes"])

                return reaction

        else:
            raise AttributeError("Cannot parse given data. Contact maintainers")


def get_response(
    query: str, model_id: Optional[str] = None
) -> tuple[str, requests.Response]:
    """
    Request different servers to see if there a valid response and returns the
    database name with the response

    model_id is a BIGG-specific argument.

    The content type must be xml, json or text
    """
    if not model_id:
        model_id = "universal"

    subdatabase = ""

    biocyc_credentials = False

    try:
        # e.g sol:META:water
        database, subdatabase, identifier = query.split(":")
        url = f"{databases[database]}{subdatabase}:{identifier}"

    except ValueError:
        database, identifier = query.split(":")
        encoded_id = urllib.parse.quote(identifier, safe="")

        if database == "KEGG":
            url = f"{databases['KEGG']}{encoded_id}"
        else:
            url = f"{databases['BIOCYC']}{database}:{encoded_id}"
            biocyc_credentials = True

    if database == "BIGG":
        query = query.removeprefix("BIGG:")
        response, _ = bigg.find_url(model_id, query)
        response.raise_for_status()

        # This extra attribute is later used to save the file in their
        # corresponding directories
        response.extra = Path(model_id)  # type: ignore

        return database, response

    s = requests.Session()
    if biocyc_credentials:
        # FIXME: find a better way
        user, pwd = cmod_utils.get_credentials(
            Path.cwd().joinpath("credentials.txt")
        )
        s.post(
            "https://websvc.biocyc.org/credentials/login/",
            data={"email": user, "password": pwd},
        )

    response = s.get(url)
    s.close()
    response.raise_for_status()

    # Not a valid content-type to process
    header = response.headers.get("Content-Type", "").lower()

    if response.text[:300].find("Subscription") != -1 and "html" in header:
        raise requests.HTTPError(
            f"Cannot retrieve data from '{url}'. It is possible that a "
            "subscription is required. Try using 'META'"
        )

    return database, response


def file_to_Data_class(
    identifier: str, filename: Path, genome: Optional[str]
) -> Data:
    """
    Creates a Data object from given file.

    Args:
        identifier (str): Identifier of the object. It should be included in
            the file itself
        filename (Path): Location of the file
        genome (Optional[str]): BIGG-specific argument. Genome involved

    Returns:
        Data
    """
    f = open(filename, "r")
    text = f.read()
    f.close()

    suffix = filename.suffix
    parent = filename.parent.name
    extra_dir = filename.parents[1].name

    if suffix == ".json":
        return Data.from_json(identifier, json.loads(text), extra_dir, filename)

    elif suffix == ".xml":
        if extra_dir == "PMN":
            parent = f"{extra_dir}:{parent}"
        # NOTE: Remove for v2.0.0
        elif extra_dir == "SOL":
            parent = f"{extra_dir}:{parent}"
            warnings.warn(
                "Database Solcyc is being deprecated for next version",
                DeprecationWarning,
            )
        return Data.from_xml(identifier, et.fromstring(text), parent, filename)

    elif suffix == ".txt":
        if not genome:
            genome = ""
        return Data.from_text(identifier, text, parent, filename, genome)

    else:
        raise AttributeError("Cannot parse given content type")


def write(name: str, directory: Path, response: requests.Response) -> Path:
    """
    Finds the content type and writes the data into disk and returns the Path
    of the new file
    """

    header = response.headers.get("Content-Type", "").lower()

    if "json" in header:
        prefix = ".json"

    elif "xml" in header:
        prefix = ".xml"

    elif "text" in header:
        prefix = ".txt"

    else:
        raise AttributeError("Cannot parse given content type")

    filename = directory.joinpath(name + prefix)

    with open(file=filename, mode="w+") as file:
        file.write(response.text)

    return filename


def get_files(directory: Path, identifier: str):
    """
    Returns a Generator with the possible files given a specific identifier
    """
    for item in directory.iterdir():
        for pattern in EXTENSIONS:
            if not item.match(identifier + pattern):
                continue
            yield item


def get_data(
    identifier: str,
    directory: Union[str, Path],
    database: Optional[str],
    model_id: Optional[str] = None,
    genome: Optional[str] = None,
) -> Data:
    """
    Retrieves the Data for given identifier. This function either retrieves
    from the server of the databases or locally.

    Args:
        identifier (str): Name of the object to retrieve
        dictionary (str or Path): Location of the files to retrieve or store
        database (Optional[str]): Name of the database. Check
            cobramod.retrieval.available_databases for more information
        model_id (Optional[str]): BIGG-specific argument. Name of the model to
            retrieve information
        genome: (Optional[str]): Name of the genome to retrieve

    Returns:
        Data
    """
    if isinstance(directory, str):
        directory = Path(directory).absolute()

    main_dir = directory

    # Biocyc db families
    family = ""
    if database is not None and database.find(":") != -1:
        family, database = (i.rstrip().strip() for i in database.split(":"))
        family = family.upper()

        directory = directory.joinpath(family)

    if not directory.exists():
        directory.mkdir()

    # Try first locally and the query databases
    if not database:
        try:
            filename = next(directory.rglob(identifier + "*"))
            response_database = filename.parent.name.upper()

        except StopIteration:
            response_database, response = get_response(identifier)

            if response_database == "bigg":
                extra: Path = getattr(response, "extra")

                if not extra:
                    raise AttributeError(
                        "The 'extra' attribute was not found in the response"
                    )
                directory = directory.joinpath(extra)

            filename = write(identifier, directory, response)
    else:
        # Create dir if needed
        if not directory.joinpath(database).exists():
            directory.joinpath(database).mkdir()

        try:
            if model_id:
                if not directory.joinpath(database, model_id).exists():
                    directory.joinpath(database, model_id).mkdir()

                filename = next(
                    get_files(
                        directory.joinpath(database, model_id), identifier
                    )
                )
                response_database = "BIGG"

            else:
                filename = next(
                    get_files(directory.joinpath(database), identifier)
                )
                response_database = filename.parent.name.upper()

        except StopIteration:
            if family:
                query = f"{family}:{database}:{identifier}"

            else:
                query = f"{database}:{identifier}"

            response_database, response = get_response(query, model_id)

            if database == "KEGG":
                kegg.retrieve_kegg_genes(directory, identifier)

            if database != "KEGG" and database != "BIGG":
                if not family:
                    biocyc.retrieve_gene_information(
                        directory, identifier, database
                    )

                elif family == "PMN":
                    plantcyc.retrieve_gene_information(
                        directory, identifier, database
                    )
                # NOTE: Deprecation for 2.0.0
                elif family == "SOL":
                    warnings.warn(
                        "Database Solcyc is being deprecated for next version",
                        DeprecationWarning,
                    )
                    solcyc.retrieve_gene_information(
                        directory, identifier, database
                    )

            if response_database == "BIGG":
                extra: Path = getattr(response, "extra")

                if not extra:
                    raise AttributeError(
                        "The 'extra' attribute was not found in the response"
                    )
                directory = directory.joinpath("BIGG", extra)

                if not directory.exists():
                    directory.mkdir()

            else:
                directory = directory.joinpath(database)

            filename = write(identifier, directory, response)

    data = file_to_Data_class(identifier, filename, genome)

    if family:
        response_database = f"{family}:{database}"

    db_configuration.check_database_version(
        main_dir, response_database, data.version
    )
    return data


def build_reaction_from_str(
    model: cobra_core.Model,
    reaction: cobra_core.Reaction,
    reaction_str: str,
    directory: Path,
    database: Optional[str],
    model_id: Optional[str],
    replacement: dict[str, str],
):
    """
    Populates the reaction with Metabolites from given reaction string.
    Metabolites are retrieved from the servers or locally

    Args:
        model (cobra.core.Model): Model to search for metabolites
        reaction (cobra.core.Reaction): Empty reaction to populate
        reaction_str (str): valid reaction string
        directory (Path): Path for the location of the directory where the
            information is stored
        database (str, optional): Name of the database. Check
            cobramod.retrieval.available_databases for more information
        model_id (str, optional): BIGG-specific argument. Name of the model to
            retrieve information
        genome: (Optional[str]): Name of the genome to retrieve
        replacement (dict[str, str]): The dictionary represents the metabolites
            to replace. The key is the identifier of the object to replaced and
            the value is the new identifier to use instead
    """
    position = cmod_utils.get_arrow_position(reaction_str)
    arrow = reaction_str[position : position + 3]
    reaction.bounds = cmod_utils.ARROWS[arrow]

    compounds: dict[str, float] = {}
    # Reactants
    for pair in reaction_str[:position].rstrip().strip().split("+"):
        try:
            coefficient, metabolite = pair.rstrip().strip().split(" ")
        except ValueError:
            coefficient = 1
            metabolite = pair.rstrip().strip()

            # Empty part
            if metabolite == "":
                break

        compounds[metabolite] = -1.0 * float(coefficient)

    # products
    for pair in reaction_str[position + 3 :].rstrip().strip().split("+"):
        try:
            coefficient, metabolite = pair.rstrip().strip().split(" ")
        except ValueError:
            coefficient = 1
            metabolite = pair.rstrip().strip()

            # Empty part
            if metabolite == "":
                break

        compounds[metabolite] = float(coefficient)

    # It must be a biocyc family
    # if database is not None and database.find(":") != -1:
    #     directory = directory.parent

    for identifier, coefficient in compounds.items():
        compartment = identifier[-1]

        identifier = replacement.get(identifier[:-2], identifier[:-2])
        identifier = f"{identifier}_{compartment}"

        try:
            data = get_data(identifier[:-2], directory, database, model_id)
            metabolite = creation.metabolite_from_data(data, compartment, model)

        except (requests.HTTPError, ValueError):
            metabolite = cobra_core.Metabolite(
                identifier, name=identifier, compartment=compartment
            )
            debug_log.debug(f"Curated Metabolite '{metabolite.id}' created")

        # NOTE: add logs
        if not isinstance(metabolite, cobra_core.Metabolite):
            raise TypeError("Given object is not a valid COBRApy Metabolite")

        reaction.add_metabolites({metabolite: coefficient})
        debug_log.debug(
            f"Metabolite '{metabolite.id}' added to reaction '{reaction.id}'"
        )



#!/usr/bin/env python3
import logging
import unittest
from contextlib import suppress
from pathlib import Path

import cobra.core as cobra_core
import cobramod.core.creation as cmod_core_creation
import cobramod.error as cmod_error
import cobramod.utils as ui
from cobra import __version__ as cobra_version
from cobramod import __version__ as cmod_version
from cobramod.debug import change_to_debug
from cobramod.test import textbook_kegg

change_to_debug()

dir_data = Path(__file__).resolve().parent.joinpath("data")
dir_input = Path(__file__).resolve().parent.joinpath("input")

# If data is missing, then do not test. Data should always be the same
if not dir_data.exists():
    raise NotADirectoryError("Data for the test is missing")


class UtilsTesting(unittest.TestCase):
    def test_check_imbalance(self):
        # Configuration
        test_reaction = cmod_core_creation.create_object(
            directory=dir_data,
            identifier="RXN-11414",
            database="ARA",
            compartment="c",
            # In order to catch warning
            show_imbalance=False,
        )
        if not isinstance(test_reaction, cobra_core.Reaction):
            raise TypeError("Not a valid COBRApy Reaction")

        # CASE: Showing imbalance
        with self.assertLogs(level=logging.DEBUG) as cm:
            ui.check_imbalance(
                reaction=test_reaction,
                show_imbalance=True,
                stop_imbalance=False,
            )
            self.assertIn("unbalanced", cm.output[-1])

        # CASE: Stopping at imbalance
        self.assertRaises(
            cmod_error.UnbalancedReaction,
            ui.check_imbalance,
            reaction=test_reaction,
            show_imbalance=True,
            stop_imbalance=True,
        )

    def test_read_lines(self):
        # CASE 0: Comments and blank lines
        with open(file=dir_input.joinpath("test_reading_lines.txt")) as f:
            line = list(ui.read_lines(f=f))
        self.assertEqual(len(line), 4)

    def test_create_replacement(self):
        # CASE 1: regular dictionary with comments and empty spaces
        test_dict = ui.create_replacement(
            filename=dir_input.joinpath("test_create_replacement.txt")
        )
        self.assertIn(member="fum_e", container=test_dict.keys())
        self.assertIn(member="Glucopyranose_e", container=test_dict.values())

    def test_compare_type(self):
        # CASE 1: Native classes
        self.assertTrue(expr=ui.compare_type(first=dict(), second=dict()))
        self.assertRaises(TypeError, ui.compare_type, str(), dict())
        # CASE 2: COBRApy classes
        self.assertTrue(
            expr=ui.compare_type(
                first=cobra_core.Reaction("0"), second=cobra_core.Reaction("1")
            )
        )
        self.assertRaises(
            TypeError,
            ui.compare_type,
            cobra_core.Reaction("0"),
            cobra_core.Metabolite("1"),
        )

    def test_compare_DictList(self):
        # Preparation
        test_model = textbook_kegg.copy()
        cmod_core_creation.add_reactions(
            model=test_model,
            directory=dir_data,
            database="KEGG",
            obj=["R00894,c"],
            replacement={},
        )
        # CASE 1: Different types
        test_generator = ui.compare_DictList(
            first=test_model.metabolites, second=test_model.reactions
        )
        self.assertRaises(TypeError, next, test_generator)
        # CASE 2: Regular Metabolites
        test_list = list(
            ui.compare_DictList(
                first=test_model.metabolites, second=textbook_kegg.metabolites
            )
        )
        self.assertEqual(first=len(test_list), second=3)
        # CASE 3: Regular Reactions
        test_list = list(
            ui.compare_DictList(
                first=test_model.reactions, second=textbook_kegg.reactions
            )
        )
        self.assertEqual(first=len(test_list), second=1)

    def test_save_to_file(self):
        # CASE 1: regular lines
        test_filename = dir_input.joinpath("summary.txt")
        with suppress(FileNotFoundError):
            test_filename.unlink()
        test_list = ["This should be first line", "This should second line"]
        ui.write_to_file(sequences=test_list, filename=test_filename)
        self.assertTrue(expr=test_filename.exists())
        with open(file=str(test_filename), mode="r") as e:
            self.assertEqual(first=2, second=sum(1 for _ in e))
        test_filename.unlink()

    def test_find_intersection(self):
        # CASE 1: Reactions Not found
        test_dict = {"A": "No", "B": "NotFound"}
        test_string = ui.find_intersection(
            dictlist=textbook_kegg.reactions,
            query=test_dict,
            revert=True,
        )
        self.assertIsNone(test_string)
        # CASE 1: Metabolites
        test_dict = {"A": "C00058", "B": "NotFound"}
        test_string = ui.find_intersection(
            dictlist=textbook_kegg.metabolites, query=test_dict, revert=True
        )
        self.assertEqual(first="C00058", second=test_string)
        # CASE 2: Reactions
        test_dict = {"A": "R00228", "B": "NotFound"}
        test_string = ui.find_intersection(
            dictlist=textbook_kegg.reactions, query=test_dict, revert=True
        )
        self.assertEqual(first="R00228", second=test_string)


if __name__ == "__main__":
    print(f"CobraMod version: {cmod_version}")
    print(f"COBRApy version: {cobra_version}")

    unittest.main(verbosity=2)



"""
Check if models are available
"""
import unittest
from pathlib import Path

import pkg_resources


class ModuleTest(unittest.TestCase):
    def test_directory(self):
        self.assertTrue(
            Path(
                pkg_resources.resource_filename("cobra", "data/textbook.xml.gz")
            ).exists()
        )
        self.assertTrue(
            Path(
                pkg_resources.resource_filename(
                    "cobramod", "data/textbook_biocyc.sbml"
                )
            ).exists()
        )
        self.assertTrue(
            Path(
                pkg_resources.resource_filename(
                    "cobramod", "data/textbook_kegg.sbml"
                )
            ).exists()
        )


if __name__ == "__main__":
    unittest.main(verbosity=2, failfast=True)



#!/usr/bin/env python3
"""Unit test for sub-package visualization

This module includes two TestCases:

- TestItems: Creation and behavior of JSON objects for the Escher-schema
- TestJsonDictionary: Testing the methods inside the JsonDictionary
"""
import unittest
from contextlib import suppress
from pathlib import Path
from time import sleep
from webbrowser import open as web_open

import cobramod.visualization.mapping as mp
from cobramod.error import FoundInPairError
from cobramod.visualization.converter import (
    JsonDictionary,
    Position,
    _convert_string,
)
from cobramod.visualization.items import Node, Reaction, Segment
from cobramod.visualization.pair import PairDictionary
from escher import Builder

dir_data = Path(__file__).resolve().parent.joinpath("data")
dir_input = Path(__file__).resolve().parent.joinpath("input")
# If data is missing, then do not test. Data should always be the same
if not dir_data.exists():
    raise NotADirectoryError("Data for the test is missing")


class TestItems(unittest.TestCase):
    """
    Behavior of JSON objects
    """

    def test__convert_string(self):
        # CASE 1: Simple reaction, reversible
        test_string = "C00002_c + C00033_c <=> C00227_c + G11113_c"
        test_dict = _convert_string(string=test_string)
        self.assertEqual(first=test_dict["C00002_c"], second=-1)
        self.assertEqual(first=test_dict["C00227_c"], second=1)
        # CASE 2: Simple reaction, irreversible
        test_string = "C00002_c + C00033_c <-- C00227_c + G11113_c"
        test_dict = _convert_string(string=test_string)
        self.assertEqual(first=test_dict["C00002_c"], second=-1)
        self.assertEqual(first=test_dict["C00227_c"], second=1)
        # CASE 3: Reaction with multiple coefficients.
        test_string = "C00001_c + 2 C00002_c --> C00009_c + C00080_c + G11113_c"
        test_dict = _convert_string(string=test_string)
        self.assertEqual(first=test_dict["C00002_c"], second=-2)
        self.assertEqual(first=test_dict["C00080_c"], second=1)

    def test_PairDictionary(self):
        # CASE 1a: Raise FoundInPairError, if common key.
        test_dict_a = PairDictionary(one=1)
        test_dict_b = PairDictionary(pair=test_dict_a)
        test_dict_b.set_pair(pair=test_dict_a)
        self.assertRaises(FoundInPairError, test_dict_b.update, one=1)
        # CASE 1b: Raise FoundInPairError, if given pair has common key.
        test_dict_a = PairDictionary(one=1)
        test_dict_b = PairDictionary(one=1)
        self.assertRaises(
            FoundInPairError, test_dict_b.set_pair, pair=test_dict_a
        )
        # CASE 2: setting a pair
        test_dict_a = PairDictionary(one=1)
        test_dict_b = PairDictionary()
        test_dict_a.set_pair(pair=test_dict_b)
        self.assertIs(expr1=test_dict_a.pair, expr2=test_dict_b)

    def test_Node(self):
        # CASE 0: Check instance behavior.
        test_class = Node(node_type="midmarker", x=1, y=2)
        test_class_2 = Node(node_type="midmarker", x=1, y=2)
        self.assertIsNot(expr1=test_class, expr2=test_class_2)
        # CASE 1a: Regular marker
        test_class = Node(node_type="midmarker", x=1, y=2)
        self.assertNotIn(member="x_label", container=test_class.keys())
        self.assertEqual(first=test_class["node_type"], second="midmarker")
        # CASE 1b: Regular metabolite
        test_class = Node(
            node_type="metabolite", x=1, y=2, label_y=1, label_x=2
        )
        self.assertIn(member="label_x", container=test_class.keys())
        self.assertEqual(first=test_class["node_type"], second="metabolite")

    def test_Segment(self):
        # CASE 0: Check instance behavior.
        test_class = Segment(from_node_id="1", to_node_id="2")
        test_class_2 = Segment(from_node_id="1", to_node_id="2")
        self.assertIsNot(expr1=test_class, expr2=test_class_2)
        # CASE 1a: Regular creation of Segment. b1 and b2 are not specified
        test_class = Segment(from_node_id="1", to_node_id="2")
        self.assertIn(member="from_node_id", container=test_class.keys())
        self.assertIn(member="b2", container=test_class.keys())
        # CASE 1b: b1 and b2 are specified
        test_class = Segment(
            from_node_id="1",
            to_node_id="2",
            b1={"x": 10, "y": 100},
            b2={"x": 20, "y": 200},
        )
        self.assertIn(member="x", container=test_class["b1"].keys())
        self.assertIn(member="y", container=test_class["b2"].keys())

    def test_Reaction(self):
        # CASE 0: Check instance behavior.
        test_class = Reaction(
            name="test_reaction",
            bigg_id="test_identifier",
            reversibility=True,
            label_x=100,
            label_y=200,
        )
        test_class_2 = Reaction(
            name="test_reaction",
            bigg_id="test_identifier",
            reversibility=False,
            label_x=10,
            label_y=20,
        )
        self.assertIsNot(expr1=test_class, expr2=test_class_2)
        # CASE 1a: Regular creation.
        test_class = Reaction(
            name="test_reaction",
            bigg_id="test_identifier",
            reversibility=True,
            label_x=100,
            label_y=200,
        )
        self.assertIn(member="metabolites", container=test_class.keys())
        # CASE 1b: Regular creation with a Segment
        test_class = Reaction(
            name="test_reaction",
            bigg_id="test_identifier",
            reversibility=True,
            label_x=100,
            label_y=200,
            segments={"0": Segment(from_node_id="0", to_node_id="1")},
        )
        self.assertIn(member="b1", container=test_class["segments"]["0"].keys())
        # CASE 2: method add_metabolite
        test_class.add_metabolite(bigg_id="test_metabolite", coefficient=1)
        self.assertEqual(
            first="test_metabolite",
            second=test_class["metabolites"][0]["bigg_id"],
        )


class TestJsonDictionary(unittest.TestCase):
    """
    Methods for the JsonDictionary.
    """

    def test___init__(self):
        # CASE 0: Checking behavior with two instances
        test_class = JsonDictionary()
        test_class_2 = JsonDictionary()
        test_class["reactions"]["0"] = "test_string"
        self.assertIsNot(expr1=test_class, expr2=test_class_2)
        self.assertRaises(KeyError, lambda: test_class_2["reactions"]["0"])
        # CASE 1: creation of dictionary without extra arguments.
        test_class = JsonDictionary()
        self.assertEqual(first={}, second=test_class["reactions"])
        self.assertEqual(first="", second=test_class["head"]["map_name"])

    def test__get_last_number(self):
        # CASE 1: get last number in JsonDictionary. Reactions are not included
        # but their segments
        test_class = JsonDictionary()
        test_class["reactions"]["99"] = Reaction(
            name="test_reaction",
            bigg_id="test_identifier",
            reversibility=True,
            label_x=100,
            label_y=200,
            segments={"0": Segment(from_node_id="0", to_node_id="1")},
        )
        test_class["nodes"]["1"] = Node(node_type="midmarker", x=1, y=2)
        self.assertEqual(
            first=2, second=test_class._get_last_number(item="nodes")
        )

    def test_get_column_reactions(self):
        # Preparing tests
        test_class = JsonDictionary()
        test_class._overview["R1"] = {"position": Position(row=1, column=1)}
        test_class._overview["R2"] = {"position": Position(row=2, column=1)}
        test_class._overview["R3"] = {"position": Position(row=3, column=0)}
        test_class._overview["R4"] = {"position": Position(row=4, column=1)}
        test_class._overview["R5"] = {"position": Position(row=5, column=0)}
        test_class._overview["R6"] = {"position": Position(row=2, column=0)}
        test_class._overview["R7"] = {"position": Position(row=2, column=3)}
        # CASE 1: Retrieving columns
        test_list = test_class._get_matrix_reactions(vertical=False, position=1)
        self.assertCountEqual(first=["R1", "R2", "R4"], second=test_list)
        # CASE 2: Retrieving Row
        test_list = test_class._get_matrix_reactions(vertical=True, position=2)
        self.assertCountEqual(first=["R2", "R6", "R7"], second=test_list)

    def test__get_products(self):
        # Preparing tests
        test_class = JsonDictionary()
        test_class._overview["R1"] = {"index": "0"}
        test_class._overview["R2"] = {"index": "1"}
        test_class._overview["R3"] = {"index": "2"}
        test_class._overview["R4"] = {"index": "3"}
        test_class["reactions"] = {
            "0": {
                "bigg_id": "R1",
                "metabolites": [
                    {"bigg_id": "C01001_c", "coefficient": -1},
                    {"bigg_id": "C01002_c", "coefficient": 1},
                ],
            },
            "1": {
                "bigg_id": "R2",
                "metabolites": [
                    {"bigg_id": "C01002_c", "coefficient": -1},
                    {"bigg_id": "C02001_c", "coefficient": 1},
                ],
            },
            "2": {
                "bigg_id": "R3",
                "metabolites": [
                    {"bigg_id": "C02001_c", "coefficient": -1},
                    {"bigg_id": "C03001_c", "coefficient": 1},
                    {"bigg_id": "C03002_c", "coefficient": 1},
                ],
            },
            "3": {
                "bigg_id": "R4",
                "metabolites": [
                    {"bigg_id": "C03001_c", "coefficient": -1},
                    {"bigg_id": "C04001_c", "coefficient": 1},
                    {"bigg_id": "C04002_c", "coefficient": 1},
                ],
            },
        }
        # CASE 1: Empty list
        test_dict = test_class._get_products(reactions=[])
        self.assertFalse(test_dict)
        # CASE 1: Simple reactions
        test_dict = test_class._get_products(reactions=["R1"])
        self.assertDictEqual(d1={"R1": ["C01002_c"]}, d2=test_dict)
        # CASE 1: Mix with complex reactions
        test_dict = test_class._get_products(reactions=["R1", "R2", "R4"])
        self.assertDictEqual(
            d1={
                "R1": ["C01002_c"],
                "R2": ["C02001_c"],
                "R4": ["C04001_c", "C04002_c"],
            },
            d2=test_dict,
        )

    def test__find_shared(self):
        # Preparing tests
        test_class = JsonDictionary()
        test_class._overview["R1"] = {
            "nodes": {"C01001_c": "10", "C01002_c": "11"}
        }
        test_class._overview["R2"] = {
            "nodes": {"C01002_c": "11", "C02001_c": "21"}
        }
        test_class._overview["R3"] = {
            "nodes": {"C02001_c": "21", "C03001_c": "31"}
        }
        test_dict = {"R1": ["C01002_c"], "R2": ["C02001_c"]}
        # CASE 1: Nothing is found
        # node_number, old_reaction
        test_tuple = test_class._find_shared(
            metabolite="C01003_c", products=test_dict
        )
        self.assertFalse(test_tuple[0])
        self.assertFalse(test_tuple[1])
        # CASE 2: regular finding
        # node_number, old_reaction
        test_tuple = test_class._find_shared(
            metabolite="C01002_c", products=test_dict
        )
        self.assertEqual(first=test_tuple[0], second="11")
        self.assertEqual(first=test_tuple[1], second="R1")

    def test_add_reaction(self):
        # CASE 1: Matrix 2x2, Unrelated
        test_class = JsonDictionary()
        test_class.add_reaction(
            string="C01001_c + C01002_c --> C01003_c + C01004_c",
            identifier="R1",
            name="Reaction-1",
            row=0,
            column=0,
            vertical=False,
        )
        test_class.add_reaction(
            string="C02001_c --> C02002_c",
            identifier="R2",
            name="Reaction-2",
            row=0,
            column=1,
            vertical=False,
        )
        test_class.add_reaction(
            string="C03001_c + C03002_c_c--> C03003_c",
            identifier="R3",
            name="Reaction-3",
            row=1,
            column=0,
            vertical=False,
        )
        test_class.add_reaction(
            string="C04001_c + C04002_c --> C04003_c + C04004_c",
            identifier="R4",
            name="Reaction-4",
            row=1,
            column=1,
            vertical=False,
        )
        # Reaction "R2" before two reactions boxes in x (650 * 2)
        # Reaction "R3" before one reaction box in x (650)
        self.assertLess(a=test_class["reactions"]["1"]["label_x"], b=650 * 2)
        self.assertLess(a=test_class["reactions"]["2"]["label_x"], b=650)
        # CASE 3: Different positions. Matrix 1x4
        test_class = JsonDictionary()
        test_class.add_reaction(
            string="C01001_c + C01002_c --> C01003_c + C01004_c",
            identifier="R1",
            name="Reaction-1",
            row=0,
            column=0,
            vertical=False,
        )
        test_class.add_reaction(
            string="C01003_c --> C02001_c",
            identifier="R2",
            name="Reaction-2",
            row=0,
            column=1,
            vertical=False,
        )
        test_class.add_reaction(
            string="C02001_c + C03001_c_c--> C03002_c",
            identifier="R3",
            name="Reaction-3",
            row=0,
            column=2,
            vertical=False,
        )
        test_class.add_reaction(
            string="C03002_c + C04001_c --> C04002_c + C04003_c",
            identifier="R4",
            name="Reaction-4",
            row=0,
            column=3,
            vertical=False,
        )
        # Reaction "R2" before two reactions boxes in x (550*2)
        # Reaction "R3" before three reaction boxes in x ()
        self.assertLess(a=test_class["reactions"]["1"]["label_x"], b=550 * 2)
        self.assertLess(a=test_class["reactions"]["2"]["label_x"], b=551 * 3)
        # Shared metabolite is node "2" in R1 and R2
        for segment in test_class["reactions"]["1"]["segments"].values():
            self.assertGreaterEqual(a=int(segment["from_node_id"]), b=2)
            self.assertGreaterEqual(a=int(segment["to_node_id"]), b=2)
            self.assertLessEqual(a=int(segment["to_node_id"]), b=11)
            self.assertLessEqual(a=int(segment["to_node_id"]), b=11)
        # TODO: CASE 2 connecting rows

    def test_color_grading(self):
        test_class = JsonDictionary()
        test_class.flux_solution = {"A": -2, "B": -1, "C": 0, "D": 1, "E": 2}

        expected_scale = [
            {"type": "value", "value": 1.0, "color": "rgb(237,192,110)"},
            {"type": "value", "value": 2.0, "color": "rgb(255,165,0)"},
            {"type": "value", "value": 0, "color": "rgb(220,220,220)"},
            {"type": "value", "value": -1.0, "color": "rgb(110,174,110)"},
            {"type": "value", "value": -2.0, "color": "rgb(0,128,0)"},
        ]

        # Case 1: 2 colors with colors defined as str or via rgb values
        color = ["orange", "green"]
        test_class.color_grading(color=color)

        self.assertCountEqual(
            test_class.reaction_scale,
            expected_scale,
            msg="Calculated color scale differs from the expected one",
        )

        color = [[255, 165, 0], [0, 128, 0]]
        test_class.color_grading(color=color)

        self.assertCountEqual(
            test_class.reaction_scale,
            expected_scale,
            msg="Calculated color scale differs from the expected one",
        )

        # Case 2: 2 colors with one defined as rgb and one defined as str

        color = ["orange", [0, 128, 0]]
        test_class.color_grading(color=color)

        self.assertCountEqual(
            test_class.reaction_scale,
            expected_scale,
            msg="Calculated color scale differs from the expected one",
        )

        color = [[255, 165, 0], "green"]
        test_class.color_grading(color=color)

        self.assertCountEqual(
            test_class.reaction_scale,
            expected_scale,
            msg="Calculated color scale differs from the expected one",
        )

        # Case 3: only one color defined
        # redefine expected scale
        expected_scale = [
            {"type": "value", "value": 1.0, "color": "rgb(237,192,110)"},
            {"type": "value", "value": 2.0, "color": "rgb(255,165,0)"},
            {"type": "value", "value": 0, "color": "rgb(220,220,220)"},
            {"type": "value", "value": -1.0, "color": "rgb(220,220,220)"},
            {"type": "value", "value": -2.0, "color": "rgb(220,220,220)"},
        ]

        color = ["orange", None]
        test_class.color_grading(color=color)

        self.assertCountEqual(
            test_class.reaction_scale,
            expected_scale,
            msg="Calculated color scale differs from the expected one",
        )

        # redefine expected scale
        expected_scale = [
            {"type": "value", "value": 1.0, "color": "rgb(220,220,220)"},
            {"type": "value", "value": 2.0, "color": "rgb(220,220,220)"},
            {"type": "value", "value": 0, "color": "rgb(220,220,220)"},
            {"type": "value", "value": -1.0, "color": "rgb(110,174,110)"},
            {"type": "value", "value": -2.0, "color": "rgb(0,128,0)"},
        ]

        color = ["sdfgdfvcbv", [0, 128, 0]]
        test_class.color_grading(color=color)

        self.assertCountEqual(
            test_class.reaction_scale,
            expected_scale,
            msg="Calculated color scale differs from the expected one",
        )

        # Case 4: 2 colors and min_max defined
        # redefine expected scale
        expected_scale = [
            {"type": "value", "value": 1.0, "color": "rgb(255,165,0)"},
            {"type": "value", "value": 0, "color": "rgb(220,220,220)"},
            {"type": "value", "value": -1.0, "color": "rgb(0,128,0)"},
        ]
        color = ["orange", "green"]
        test_class.color_grading(color=color, min_max=[-1, 1])

        self.assertCountEqual(
            test_class.reaction_scale,
            expected_scale,
            msg="Calculated color scale differs from the expected one",
        )

        # Case 5: no values left due to min_max

        expected_scale = [
            {"type": "value", "value": 0, "color": "rgb(220,220,220)"}
        ]
        color = ["orange", "green"]
        test_class.color_grading(color=color, min_max=[0, 0])

        self.assertCountEqual(
            test_class.reaction_scale,
            expected_scale,
            msg="Calculated color scale differs from the expected one",
        )

        test_class.color_grading(color=color, min_max=[5, 10])
        expected_scale = [
            {"type": "value", "value": 5, "color": "rgb(237,192,110)"},
            {"type": "value", "value": 10, "color": "rgb(255,165,0)"},
        ]

        self.assertCountEqual(
            test_class.reaction_scale,
            expected_scale,
            msg="Calculated color scale differs from the expected one",
        )

        expected_scale = [
            {"color": "rgb(255,165,0)", "type": "value", "value": 0.1},
            {"color": "rgb(220,220,220)", "type": "value", "value": 0},
            {"color": "rgb(0,128,0)", "type": "value", "value": -0.1},
        ]
        test_class.color_grading(color=color, min_max=[-0.1, 0.1])

        self.assertCountEqual(
            test_class.reaction_scale,
            expected_scale,
            msg="Calculated color scale differs from the expected one",
        )

        # Case 6: values close together with and without quantile
        test_class.flux_solution = {
            "A": -2,
            "B": 0.1,
            "C": 0.2,
            "D": 0.3,
            "E": 2,
        }
        expected_scale = [
            {"color": "rgb(228,206,165)", "type": "value", "value": 0.5},
            {"color": "rgb(237,192,110)", "type": "value", "value": 1.0},
            {"color": "rgb(246,178,55)", "type": "value", "value": 1.5},
            {"color": "rgb(255,165,0)", "type": "value", "value": 2.0},
            {"color": "rgb(220,220,220)", "type": "value", "value": 0},
            {"color": "rgb(0,128,0)", "type": "value", "value": -2.0},
        ]

        test_class.color_grading(color=color, quantile=False)

        self.assertCountEqual(
            test_class.reaction_scale,
            expected_scale,
            msg="Calculated color scale differs from the expected one",
        )

        expected_scale = [
            {"color": "rgb(228,206,165)", "type": "value", "value": 0.1},
            {"color": "rgb(237,192,110)", "type": "value", "value": 0.2},
            {"color": "rgb(246,178,55)", "type": "value", "value": 0.3},
            {"color": "rgb(255,165,0)", "type": "value", "value": 2.0},
            {"color": "rgb(220,220,220)", "type": "value", "value": 0},
            {"color": "rgb(0,128,0)", "type": "value", "value": -2.0},
        ]
        test_class.color_grading(color=color, quantile=True)

        self.assertCountEqual(
            test_class.reaction_scale,
            expected_scale,
            msg="Calculated color scale differs from the expected one",
        )

        # Case 7: no flux
        test_class.flux_solution = {}

        expected_scale = [
            {"color": "rgb(220,220,220)", "type": "value", "value": 0}
        ]

        test_class.color_grading(color=color)

        self.assertCountEqual(
            test_class.reaction_scale,
            expected_scale,
            msg="Calculated color scale differs from the expected one",
        )

        test_class.color_grading(color=color, quantile=True)

        self.assertCountEqual(
            test_class.reaction_scale,
            expected_scale,
            msg="Calculated color scale differs from the expected one",
        )

    def test_json_dump(self):
        # CASE 1: Simple HTML and JSON with 4 reactions
        test_class = JsonDictionary()
        # Escher builder
        test_builder = Builder()
        # Matrix 1x4
        test_class.add_reaction(
            string="C01001_c + C01002_c --> C01003_c + C01004_c",
            identifier="R1",
            name="Reaction-1",
            row=0,
            column=0,
            vertical=False,
        )
        test_class.add_reaction(
            string="C01003_c --> C02001_c",
            identifier="R2",
            name="Reaction-2",
            row=0,
            column=1,
            vertical=False,
        )
        test_class.add_reaction(
            string="C02001_c + C03001_c_c--> C03002_c",
            identifier="R3",
            name="Reaction-3",
            row=0,
            column=2,
            vertical=False,
        )
        test_class.add_reaction(
            string="C03002_c + C04001_c --> C04002_c + C04003_c",
            identifier="R4",
            name="Reaction-4",
            row=0,
            column=3,
            vertical=False,
        )
        # Writing the JSON
        test_string = test_class.json_dump(indent=4)
        # Load the JSON and save the builder. Remove previous files.
        test_builder.map_json = test_string
        test_path = Path.cwd().joinpath("test_map.html")
        with suppress(FileNotFoundError):
            test_path.unlink()
        test_builder.save_html(str(test_path))
        self.assertTrue(expr=test_path.exists())

    def test_visualize(self):
        # NOTE: visual tests
        # Settings
        test_path = Path.cwd().joinpath("test_map.html")
        test_reactions = {
            "R1": "C01001_c + C01002_c --> C01003_c + C01004_c",
            "R2": "C01003_c --> C02001_c",
            "R3": "C02001_c + C03001_c--> C03002_c",
            "R4": "C03002_c + C04001_c --> C04002_c + C04003_c",
        }
        # CASE 1: Simple single reaction
        test_class = JsonDictionary()
        test_class.graph = {"R1": None}
        test_class.reaction_strings = test_reactions
        with suppress(FileNotFoundError):
            test_path.unlink()
        test_builder = test_class.visualize(filepath=test_path)
        web_open("pathway.html")
        sleep(0.5)
        self.assertEqual(first=test_builder.reaction_data, second=None)
        self.assertTrue(expr=test_path.exists())
        # CASE 2: visualization with Data
        test_class = JsonDictionary()
        test_class.graph = {"R1": None}
        test_class.reaction_strings = test_reactions
        with suppress(FileNotFoundError):
            test_path.unlink()
        # Setting flux
        test_flux = {"R1": 2, "R2": -1}
        test_class.flux_solution = test_flux
        test_builder = test_class.visualize(filepath=test_path)
        web_open("pathway.html")
        sleep(0.5)
        self.assertEqual(first=test_builder.reaction_data["R1"], second=2)
        self.assertTrue(expr=test_path.exists())
        # CASE 3: Unrelated reactions without data
        test_class = JsonDictionary()
        test_class.graph = {"R1": None, "R2": None, "R3": None}
        # Metabolites need to be unrelated
        test_class.reaction_strings = {
            "R1": "C01001_c + C01002_c --> C01003_c + C01004_c",
            "R2": "C02003_c --> C02001_c",
            "R3": "C03001_c + C03003_c--> C03002_c",
        }
        with suppress(FileNotFoundError):
            test_path.unlink()
        test_builder = test_class.visualize(filepath=test_path)
        web_open("pathway.html")
        sleep(0.5)
        # CASE 4: regular lineal visualization without data
        test_class = JsonDictionary()
        with suppress(FileNotFoundError):
            test_path.unlink()
        test_class.graph = {"R1": "R2", "R2": "R3", "R3": None}
        test_class.reaction_strings = {
            "R1": "C00002_c + C00009_c --> C00227_c + C00003_c",
            "R2": "C00003_c --> C00228_c",
            "R3": "C00009_c + C00228_c--> C00004_c",
        }
        test_builder = test_class.visualize(filepath=test_path)
        web_open("pathway.html")
        sleep(0.5)
        # CASE 5: Simple Branch
        test_class = JsonDictionary()
        with suppress(FileNotFoundError):
            test_path.unlink()
        test_class.graph = {
            "R1": "R2",
            "R2": ("R3", "R5"),
            "R3": "R4",
            "R4": None,
            "R5": None,
        }
        test_class.reaction_strings = {
            "R1": "C00002_c + C00009_c --> C00227_c + C00003_c",
            "R2": "C00003_c --> C00228_c",
            "R3": "C00009_c + C00228_c--> C00004_c",
            "R4": "C00004_c + C00011_c --> C00001_c + C00200_c",
            "R5": "2 C00228_c --> 4 C00021_c",
        }
        test_builder = test_class.visualize(filepath=test_path)
        web_open("pathway.html")
        sleep(0.5)
        # CASE 3a: Complex Lineal
        test_class = JsonDictionary()
        with suppress(FileNotFoundError):
            test_path.unlink()
        test_class.graph = {
            "R1": "R2",
            "R2": ("R3", "R5", "R4"),
            "R3": ("R6", "R8"),
            "R4": None,
            "R5": None,
            "R6": "R7",
            "R7": "R10",
            "R8": ("R9", "R11"),
            "R9": None,
            "R10": "R14",
            "R11": ("R12", "R13"),
            "R12": None,
            "R13": None,
            "R14": None,
        }
        test_class.reaction_strings = {
            "R1": "C01001_c + C01002_c --> C01003_c + C01004_c",
            "R2": "C01003_c --> C02001_c",
            "R3": "C02001_c + C03001_c--> C03002_c",
            "R4": "C02001_c+ C04001_c --> C04002_c + C04003_c",
            "R5": "2 C02001_c --> 4 C05001_c",
            "R6": "C03002_c --> C06001_c",
            "R7": "3 C06001_c --> 6 C07001_c",
            "R8": "C03002_c --> 4 C08001_c",
            "R9": "2 C08001_c --> C09001_c",
            "R10": "2 C07001_c --> 3 C10001_c",
            "R11": "2 C08001_c --> 4 C11001_c",
            "R12": "2 C11001_c --> C12001_c",
            "R13": "C11001_c --> C13001_c",
            "R14": "C10001_c --> C14001_c",
        }
        test_builder = test_class.visualize(filepath=test_path)
        web_open("pathway.html")
        sleep(0.5)
        # CASE 4: Module from KEGG
        test_path = Path.cwd().joinpath("test_map.html")
        test_class = JsonDictionary()
        with suppress(FileNotFoundError):
            test_path.unlink()
        test_class.graph = {
            "R0": "R3",
            "R1": None,
            "R2": None,
            "R3": ("R4", "R5"),
            "R4": "R6",
            "R5": None,
            "R6": "R7",
            "R7": ("R8", "R9"),
            "R8": "R10",
            "R9": None,
            "R10": "R11",
            "R11": "R12",
            "R12": "R13",
            "R13": "R14",
            "R14": None,
        }
        test_class.reaction_strings = {
            "R0": "C00001_c --> C00002_c",
            "R1": "C01001_c + C01002_c --> C00002_c + C01004_c",
            "R2": "C02001_c --> C00002_c",
            "R3": "C00002_c + C03001_c--> C03002_c",
            "R4": "C03002_c+ C04001_c --> C04002_c + C04003_c",
            "R5": "2 C03002_c --> 4 C05001_c",
            "R6": "C04002_c --> C06001_c",
            "R7": "3 C06001_c --> 6 C07001_c",
            "R8": "C07001_c --> 4 C08001_c",
            "R9": "2 C07001_c --> C09001_c",
            "R10": "2 C08001_c --> 3 C10001_c",
            "R11": "2 C10001_c --> 4 C11001_c",
            "R12": "2 C11001_c --> C12001_c",
            "R13": "C12001_c --> C13001_c",
            "R14": "C13001_c --> C14001_c",
        }
        test_builder = test_class.visualize(filepath=test_path)
        web_open("pathway.html")
        sleep(0.5)
        with suppress(FileNotFoundError):
            test_path.unlink()
        # with color
        test_class.flux_solution = {
            "R0": 10,
            "R1": 9,
            "R2": 8,
            "R3": 7,
            "R4": 6,
            "R5": 5,
            "R6": 3,
            "R7": 3,
            "R8": 3,
            "R9": 5,
            "R10": 5,
            "R11": 4,
            "R12": 3,
            "R13": 6,
            "R14": 7,
        }
        test_builder = test_class.visualize(
            filepath=test_path, color=["orange", "green"]
        )
        web_open("pathway.html")
        sleep(0.5)

    def test_visualize_vertical(self):
        test_path = Path.cwd().joinpath("test_map.html")
        # CASE 1: Unrelated reactions without data
        test_class = JsonDictionary()
        test_class.graph = {"R1": None, "R2": None, "R3": None}
        # Metabolites need to be unrelated
        test_class.reaction_strings = {
            "R1": "C01001_c + C01002_c --> C01003_c + C01004_c",
            "R2": "C02003_c --> C02001_c",
            "R3": "C03001_c + C03003_c--> C03002_c",
        }
        with suppress(FileNotFoundError):
            test_path.unlink()
        test_class.visualize(filepath=test_path, vertical=True)
        web_open("pathway.html")
        sleep(0.5)

        # Color test
        with suppress(FileNotFoundError):
            test_path.unlink()
        test_class.flux_solution = {"R1": -4, "R2": -2, "R3": 0}
        test_class.visualize(
            filepath=test_path, vertical=True, color=["orange", "green"]
        )
        web_open("pathway.html")
        sleep(0.5)
        # CASE 2: Complex Lineal
        test_class = JsonDictionary()
        with suppress(FileNotFoundError):
            test_path.unlink()
        test_class.graph = {
            "R1": "R2",
            "R2": ("R3", "R5", "R4"),
            "R3": ("R6", "R8"),
            "R4": None,
            "R5": None,
            "R6": "R7",
            "R7": "R10",
            "R8": ("R9", "R11"),
            "R9": None,
            "R10": "R14",
            "R11": ("R12", "R13"),
            "R12": None,
            "R13": None,
            "R14": None,
        }
        test_class.reaction_strings = {
            "R1": "C01001_c + C01002_c --> C01003_c + C01004_c",
            "R2": "C01003_c --> C02001_c",
            "R3": "C02001_c + C03001_c--> C03002_c",
            "R4": "C02001_c+ C04001_c --> C04002_c + C04003_c",
            "R5": "2 C02001_c --> 4 C05001_c",
            "R6": "C03002_c --> C06001_c",
            "R7": "3 C06001_c --> 6 C07001_c",
            "R8": "C03002_c --> 4 C08001_c",
            "R9": "2 C08001_c --> C09001_c",
            "R10": "2 C07001_c --> 3 C10001_c",
            "R11": "2 C08001_c --> 4 C11001_c",
            "R12": "2 C11001_c --> C12001_c",
            "R13": "C11001_c --> C13001_c",
            "R14": "C10001_c --> C14001_c",
        }
        test_class.visualize(filepath=test_path, vertical=True)


class TestMapping(unittest.TestCase):
    """
    This TestCase checks if the mapping for the visualization has a normal
    behavior
    """

    def test_child_map(self):
        test_dict = {
            "R1": "R2",
            "R2": ("R3", "R5", "R4"),
            "R3": ("R6", "R8"),
            "R4": None,
            "R5": None,
            "R6": "R7",
            "R7": "R10",
            "R8": ("R9", "R11"),
            "R9": None,
            "R10": "R14",
            "R11": ("R12", "R13"),
            "R12": None,
            "R13": None,
            "R14": None,
        }
        test_list = [
            ["R1", "R2", "R3", "R6", "R7", "R10", "R14"],
            ["R8", "R11", "R12"],
            ["R4"],
            ["R5"],
            ["R9"],
            ["R13"],
        ]
        test_answer = mp.child_map(mapping=test_list, dictionary=test_dict)
        self.assertCountEqual(first=test_answer["1"], second=["4", "5"])
        self.assertCountEqual(first=test_answer["0"], second=["2", "3", "1"])

    def test_unformatted_matrix(self):
        # CASE 0a: Unrelated items
        test_dict = {"R1": None, "R2": None, "R3": None}
        test_matrix = mp.unformatted_matrix(graph=test_dict)
        self.assertIn(member=["R1"], container=test_matrix)
        self.assertIn(member=["R2"], container=test_matrix)
        self.assertIn(member=["R3"], container=test_matrix)
        # CASE 0b: Single unrelated item
        test_dict = {"R1": "R2", "R2": None, "R3": None}
        test_matrix = mp.unformatted_matrix(graph=test_dict)
        self.assertIn(member=["R1", "R2"], container=test_matrix)
        self.assertIn(member=["R3"], container=test_matrix)
        # CASE 1a: Simple Lineal
        test_dict = {"R1": "R2", "R2": "R3", "R3": None}
        test_matrix = mp.unformatted_matrix(graph=test_dict)
        self.assertIn(member=["R1", "R2", "R3"], container=test_matrix)
        # CASE 1b: Simple Lineal
        test_dict = {"R1": ("R2", "R4"), "R2": "R3", "R3": None, "R4": None}
        test_matrix = mp.unformatted_matrix(graph=test_dict)
        self.assertIn(member=["R1", "R2", "R3"], container=test_matrix)
        self.assertIn(member=[0, "R4"], container=test_matrix)
        # CASE 2: Simple Cyclic
        test_dict = {"R1": "R2", "R2": "R3", "R3": "R1"}
        test_matrix = mp.unformatted_matrix(graph=test_dict)
        self.assertCountEqual(first=["R1", "R2", "R3"], second=test_matrix[0])
        # CASE 3a: Complex Lineal
        test_dict = {
            "R1": "R2",
            "R2": ("R3", "R5", "R4"),
            "R3": ("R6", "R8"),
            "R4": None,
            "R5": None,
            "R6": "R7",
            "R7": "R10",
            "R8": ("R9", "R11"),
            "R9": None,
            "R10": "R14",
            "R11": ("R12", "R13"),
            "R12": None,
            "R13": None,
            "R14": None,
        }
        test_matrix = mp.unformatted_matrix(graph=test_dict)
        self.assertIn(member=[0, 0, "R5"], container=test_matrix)
        self.assertIn(member=[0, 0, "R4"], container=test_matrix)
        self.assertIn(
            member=[0, 0, 0, "R8", "R11", "R12"], container=test_matrix
        )
        # CASE 3b: Complex Lineal
        test_dict = {
            "R0": "R1",
            "R1": ("R2", "R7"),
            "R2": "R3",
            "R3": "R4",
            "R4": "R5",
            "R5": ("R6", "R9"),
            "R6": "R12",
            "R7": "R8",
            "R8": ("R10", "R11"),
            "R9": None,
            "R10": None,
            "R11": None,
            "R12": None,
        }
        test_matrix = mp.unformatted_matrix(graph=test_dict)
        self.assertIn(member=[0, 0, 0, 0, 0, 0, "R9"], container=test_matrix)
        self.assertIn(member=[0, 0, 0, 0, "R11"], container=test_matrix)

    def test_fill_matrix(self):
        # CASE 1: Normal use
        test_matrix = [["R1", "R2", "R3", "R4", "R5"], [0, 0, "R6"], [0, "R7"]]
        mp.fill_matrix(test_matrix, length=5)
        self.assertListEqual(list1=test_matrix[1], list2=[0, 0, "R6", 0, 0])
        self.assertListEqual(list1=test_matrix[2], list2=[0, "R7", 0, 0, 0])

    def test_format_matrix(self):
        # CASE 0a: only filling
        test_matrix = [["R1", "R2", "R3", "R4", "R5"], [0, 0, "R6"]]
        test_answer = mp.format_matrix(matrix=test_matrix.copy(), max_length=5)
        test_values = set()
        for row in test_answer:
            for i in row:
                if i != 0:
                    test_values.add(i)
        self.assertCountEqual(
            first=["R1", "R2", "R3", "R4", "R5", "R6"], second=test_values
        )
        self.assertEqual(first=len(test_answer), second=2)
        self.assertListEqual(list1=test_answer[1], list2=[0, 0, "R6", 0, 0])
        # CASE 0b: Unrelated
        test_matrix = [["R1"], ["R2"], ["R3"]]
        test_answer = mp.format_matrix(matrix=test_matrix.copy(), max_length=1)
        self.assertEqual(first=test_matrix, second=test_answer)
        # CASE 1: Simple modification
        test_matrix = [
            ["R0", "R1", "R2", "R3", "R4", "R5", "R6", "R12"],
            [0, 0, "R7", "R8", "R10"],
            [0, 0, 0, 0, 0, 0, "R9"],
            [0, 0, 0, 0, "R11"],
        ]
        test_answer = mp.format_matrix(matrix=test_matrix.copy(), max_length=8)
        test_values = set()
        for row in test_answer:
            for i in row:
                if i != 0:
                    test_values.add(i)
        self.assertCountEqual(
            first=[
                "R0",
                "R1",
                "R2",
                "R3",
                "R4",
                "R5",
                "R6",
                "R7",
                "R8",
                "R9",
                "R10",
                "R11",
                "R12",
            ],
            second=test_values,
        )
        self.assertEqual(first=len(test_answer), second=3)
        self.assertListEqual(
            list1=test_answer[1], list2=[0, 0, "R7", "R8", "R10", 0, "R9", 0]
        )
        # CASE 2: Complex modification
        test_matrix = [
            ["R1", "R2", "R3", "R6", "R7", "R10", "R14"],
            [0, 0, 0, "R8", "R11", "R12"],
            [0, 0, "R4"],
            [0, 0, "R5"],
            [0, 0, 0, 0, "R9"],
            [0, 0, 0, 0, 0, "R13"],
        ]
        test_answer = mp.format_matrix(matrix=test_matrix.copy(), max_length=7)
        test_values = set()
        for row in test_answer:
            for i in row:
                if i != 0:
                    test_values.add(i)
        self.assertCountEqual(
            first=[
                "R1",
                "R2",
                "R3",
                "R4",
                "R5",
                "R6",
                "R7",
                "R8",
                "R9",
                "R10",
                "R11",
                "R12",
                "R13",
                "R14",
            ],
            second=test_values,
        )
        self.assertIn(
            member=[0, 0, 0, 0, "R9", "R13", 0], container=test_answer
        )

    def test_get_all_values(self):
        # CASE 1: Simple dictionary
        test_dict = {"R1": "R2", "R2": "R3", "R3": None}
        test_set = mp.get_all_values(dictionary=test_dict, keys=["R1"])
        self.assertCountEqual(first=test_set, second={"R2"})
        # CASE 2: Simple dictionary, multiple keys
        test_dict = {
            "R0": "R1",
            "R1": ("R2", "R7"),
            "R2": "R3",
            "R3": "R4",
            "R4": "R5",
            "R5": None,
        }
        test_set = mp.get_all_values(
            dictionary=test_dict, keys=["R1", "R5", "R3"]
        )
        self.assertCountEqual(first=test_set, second={"R2", "R7", "R4"})

    def test_transpose(self):
        # CASE 0: Simple Matrix m*m
        test_matrix = [["R1", "R2"], ["R3", "R4"]]
        test_answer = mp.transpose(matrix=test_matrix)
        self.assertEqual(first=[["R1", "R3"], ["R2", "R4"]], second=test_answer)
        # CASE 1: Different dimensions m*n
        test_matrix = [["R1", "R2", "R3", "R4"], [0, 0, "R5", 0]]
        test_answer = mp.transpose(matrix=test_matrix)
        self.assertEqual(
            first=[["R1", 0], ["R2", 0], ["R3", "R5"], ["R4", 0]],
            second=test_answer,
        )
        # CASE 2: Matrix with multiple reactions
        test_matrix = [
            ["R0", "R1", "R2", "R3", "R4", "R5", "R6", "R12"],
            [0, 0, "R7", "R8", "R10", 0, "R9", 0],
            [0, 0, 0, 0, "R11", 0, 0, 0],
        ]
        test_answer = mp.transpose(matrix=test_matrix)
        self.assertEqual(
            first=[
                ["R0", 0, 0],
                ["R1", 0, 0],
                ["R2", "R7", 0],
                ["R3", "R8", 0],
                ["R4", "R10", "R11"],
                ["R5", 0, 0],
                ["R6", "R9", 0],
                ["R12", 0, 0],
            ],
            second=test_answer,
        )


if __name__ == "__main__":
    unittest.main(verbosity=2)
